"""
This module contains a Helper class to make config files.

The processing config is still evolving and this class and its methods may change.

Development Notes:
 - Initially, we only supported EMTF style band specification, but this should be updated
     to allow logarithmically spaced bands of constant Q.

"""

import pathlib
from typing import Optional, Union

from loguru import logger
from mt_metadata.processing.window import Window
from mth5.processing.kernel_dataset import KernelDataset

from aurora.config import BANDS_DEFAULT_FILE
from aurora.config.metadata import Processing
from aurora.sandbox.io_helpers.emtf_band_setup import EMTFBandSetupFile


SUPPORTED_BAND_SPECIFICATION_STYLES = ["EMTF", "band_edges"]


class ConfigCreator:
    def __init__(
        self,
        emtf_band_file: Optional[Union[str, pathlib.Path, None]] = None,
        band_edges: Optional[Union[dict, None]] = None,
    ):
        """
        Constructor

        Parameters
        ----------
        emtf_band_file: Optional[Union[str, pathlib.Path, None]]
            Allows the specification of an EMTF "band setup file" for defining frequency bands.
        band_edges: dict
            Keys are integers corresponding to decimation level.  Values are numpy arrays.
            Numpy arrays are one-row-per-band.  Array shape n_bands x 2. array[i_band,0] is the
            lower edge of band, and array[i_band,0] is the upper.

        """
        self._emtf_band_file = emtf_band_file
        self._band_edges = band_edges
        self._band_specification_style = None

    # TODO: Delete or update with config info ... currently kernel_dataset.processing_id is used.
    # def processing_id(self, kernel_dataset):
    #     """
    #     Generates a string id label for the processing config: WIP.

    #     In the past, we used f"{local}-{remote}" or  f"{local}-{run_id}"
    #     Neither of these is sufficiently unique.  In fact, they only describe the
    #     dataset, and not the processing config.  It is difficult to see how to make a
    #     comprehensive, unique id without it being very long or involving hash
    #     functions.

    #     For now, will try to use {local}-{local_runs}-{remote}-{remote_runs},
    #     which at least describes the dataset, then a string can be generated by the
    #     config and appended if needed.

    #     Parameters
    #     ----------
    #     kernel_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
    #         An object that defines the data to be processed.

    #     Returns
    #     -------
    #     id: str
    #         A label for the processing config.
    #     """
    #     id = f"{kernel_dataset.local_station_id}-{kernel_dataset.remote_station_id}"
    #     return id

    @property
    def band_specification_style(self):
        """return a description of the scheme used to define the bands."""
        return self._band_specification_style

    @band_specification_style.setter
    def band_specification_style(self, value: str) -> None:
        """
        Sets the band_specification_style

        Parameters
        ----------
        value: str
            The label for the scheme used to define the bands.

        Returns
        -------

        """
        if value not in SUPPORTED_BAND_SPECIFICATION_STYLES:
            msg = f"Won't set band specification style to unrecognized value {value}"
            logger.warning(msg)
            raise NotImplementedError(msg)
        else:
            self._band_specification_style = value

    def determine_band_specification_style(self) -> None:
        """
            Tries to identify which scheme was used to define the bands

        TODO: Should emtf_band_file path be stored in config to support reproducibility?

        """

        if (self._emtf_band_file is None) & (self._band_edges is None):
            msg = "Bands not defined; setting to EMTF BANDS_DEFAULT_FILE"
            logger.info(msg)
            self._emtf_band_file = BANDS_DEFAULT_FILE
            self._band_specification_style = "EMTF"
        elif (self._emtf_band_file is not None) & (self._band_edges is not None):
            msg = "Bands defined twice, and possibly inconsistently"
            logger.error(msg)
            raise ValueError(msg)
        elif self._band_edges is not None:
            self._band_specification_style = "band_edges"
        elif self._emtf_band_file is not None:
            self._band_specification_style = "EMTF"

    def create_from_kernel_dataset(
        self,
        kernel_dataset: KernelDataset,
        input_channels: Optional[list] = None,
        output_channels: Optional[list] = None,
        remote_channels: Optional[list] = None,
        estimator: Optional[str] = None,
        emtf_band_file: Optional[Union[str, pathlib.Path]] = None,
        band_edges: Optional[dict] = None,
        decimation_factors: Optional[list] = None,
        num_samples_window: Optional[int] = None,
        **kwargs,
    ) -> Processing:
        """
        Creates a processing config from a kernel dataset.

        TODO: Make this a method of kernel_dataset.

        **Development Notes:**

        1. The number of decimation levels must be defined either by:

        - decimation_factors argument (normally accompanied by a bands_dict)
        - number of decimations implied by EMTF band setup file.

        Theoretically, you could also use the number of decimations implied by bands_dict but this is sloppy, because it would assume the decimation factor.

        2. 2024-12-29 Added setting of decimation_obj.stft.per_window_detrend_type = "linear"

        This makes tests pass following a refactoring of mt_metadata.  Could use more testing.

        Parameters
        ----------
        kernel_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
            An object that defines the data to be processed.
        input_channels: list
            List of the input channels that will be used in TF estimation (usually "hx", "hy")
        output_channels: list
            List of the output channels that will be estimated by TF (usually "ex", "ey", "hz")
        remote_channels: list
            List of the remote reference channels (usually "hx", "hy" at remote site)
        estimator:  Optional[Union[str, None]]
            The name of the regression estimator to use for TF estimation.
        emtf_band_file: Optional[Union[str, pathlib.Path, None]]
            The emtf and setup file if used.
        band_edges: Optional[Union[dict, None]]
            The band edges if emtf_band_file not used
        decimation_factors: Optional[Union[list, None]]
            List of decimation factors, normally [1, 4, 4, 4, ... 4]
        num_samples_window: Optional[Union[int, None]]
            The size of the window (usually for FFT)

        **kwargs:
            Additional keyword arguments passed to Processing constructor. Could contain:

            - save_fcs: bool
                If True, save Fourier coefficients during processing.
            - save_fcs_type: str
                File type for saving Fourier coefficients. Options are "h5" or "csv".

        Returns
        -------
        processing_obj: aurora.config.metadata.processing.Processing
            Object storing the processing parameters.
        """

        processing_obj = Processing(id=kernel_dataset.processing_id)  # , **kwargs)

        # pack station and run info into processing object
        processing_obj.stations.from_dataset_dataframe(kernel_dataset.df)

        # Unpack optional arguments
        self._emtf_band_file = emtf_band_file
        self._band_edges = band_edges
        decimation_factors = decimation_factors
        num_samples_window = num_samples_window

        # Determine if band_setup or edges dict is to be used for bands
        self.determine_band_specification_style()

        # Set Frequency Bands
        if self.band_specification_style == "EMTF":
            emtf_band_setup_file = EMTFBandSetupFile(
                filepath=self._emtf_band_file,
                sample_rate=kernel_dataset.sample_rate,
            )
            num_decimations = emtf_band_setup_file.num_decimation_levels
            # Assign optional arguments if they have not been passed
            if decimation_factors is None:
                # set default values to EMTF default values [1, 4, 4, 4, ..., 4]
                decimation_factors = num_decimations * [4]
                decimation_factors[0] = 1
            if num_samples_window is None:
                default_window = Window()
                num_samples_window = num_decimations * [default_window.num_samples]
            elif isinstance(num_samples_window, int):
                num_samples_window = num_decimations * [num_samples_window]
            # now you can define the frequency bands
            band_edges = emtf_band_setup_file.compute_band_edges(
                decimation_factors, num_samples_window
            )
            self._band_edges = band_edges

        processing_obj.assign_bands(
            self._band_edges,
            kernel_dataset.sample_rate,
            decimation_factors,
            num_samples_window,
        )
        processing_obj.band_specification_style = self.band_specification_style
        if self.band_specification_style == "EMTF":
            processing_obj.band_setup_file = str(self._emtf_band_file)
        for key, decimation_obj in processing_obj.decimations_dict.items():
            if input_channels is None:
                decimation_obj.input_channels = kernel_dataset.input_channels
            else:
                decimation_obj.input_channels = input_channels

            if output_channels is None:
                decimation_obj.output_channels = kernel_dataset.output_channels
            else:
                decimation_obj.output_channels = output_channels

            if remote_channels is None:
                if kernel_dataset.remote_channels is not None:
                    decimation_obj.reference_channels = kernel_dataset.remote_channels

            if num_samples_window is not None:
                decimation_obj.stft.window.num_samples = num_samples_window[key]

            if kwargs.get("save_fcs", False):
                decimation_obj.save_fcs = True
                decimation_obj.save_fcs_type = kwargs.get("save_fcs_type", "h5")

            # set estimator if provided as kwarg
            if estimator:
                try:
                    decimation_obj.estimator.engine = estimator["engine"]
                except KeyError:
                    pass
            decimation_obj.stft.per_window_detrend_type = "linear"

        return processing_obj



"""Pytest translation of test_stft_methods_agree.py

This test confirms that the internal aurora stft method returns the same
array as scipy.signal.spectrogram
"""

import numpy as np
from mth5.helpers import close_open_files
from mth5.mth5 import MTH5
from mth5.processing import KernelDataset, RunSummary
from mth5.processing.spectre.stft import run_ts_to_stft_scipy

from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.test_utils.synthetic.make_processing_configs import create_test_run_config
from aurora.time_series.spectrogram_helpers import run_ts_to_stft


def test_stft_methods_agree(worker_safe_test1_h5, synthetic_test_paths):
    """Test that aurora STFT and scipy STFT produce identical results.

    The answer is "mostly yes", under two conditions:
    1. scipy.signal.spectrogram does not innately support an extra linear
       detrending to be applied _after_ tapering.
    2. We do not wish to apply "per-segment" prewhitening as is done in some
       variations of EMTF.

    Excluding these, we get numerically identical results.
    """
    close_open_files()
    mth5_path = worker_safe_test1_h5

    run_summary = RunSummary()
    run_summary.from_mth5s([mth5_path])
    tfk_dataset = KernelDataset()
    station_id = "test1"
    run_id = "001"
    tfk_dataset.from_run_summary(run_summary, station_id)

    processing_config = create_test_run_config(station_id, tfk_dataset)

    mth5_obj = MTH5(file_version="0.1.0")
    mth5_obj.open_mth5(mth5_path, mode="a")

    for dec_level_id, dec_config in enumerate(processing_config.decimations):
        if dec_level_id == 0:
            run_obj = mth5_obj.get_run(station_id, run_id, survey=None)
            run_ts = run_obj.to_runts(start=None, end=None)
            local_run_xrts = run_ts.dataset
        else:
            local_run_xrts = prototype_decimate(dec_config.decimation, local_run_xrts)

        dec_config.stft.per_window_detrend_type = "constant"
        local_spectrogram = run_ts_to_stft(dec_config, local_run_xrts)
        local_spectrogram2 = run_ts_to_stft_scipy(dec_config, local_run_xrts)
        stft_difference = (
            local_spectrogram.dataset - local_spectrogram2.dataset
        ).to_array()

        # drop dc term
        stft_difference = stft_difference.where(
            stft_difference.frequency > 0, drop=True
        )

        assert np.isclose(stft_difference, 0).all()



"""
    This module contains helper functions for making mth5 from FDSN clients.

"""
import pathlib
from pathlib import Path
from typing import Optional, Union

import obspy
from loguru import logger
from mt_metadata.timeseries.stationxml import XMLInventoryMTExperiment
from mth5.timeseries import RunTS
from mth5.utils.helpers import initialize_mth5

from aurora.sandbox.obspy_helpers import (
    align_streams,
    make_channel_labels_fdsn_compliant,
    trim_streams_to_common_timestamps,
)
from aurora.sandbox.triage_metadata import (
    triage_missing_coil_hollister,
    triage_mt_units_electric_field,
    triage_mt_units_magnetic_field,
)


def create_from_server_multistation(
    fdsn_dataset,
    target_folder: Optional[pathlib.Path] = Path(),
    run_id: Optional[str] = "001",
    force_align_streams: Optional[bool] = True,
    triage_units: Optional[Union[list, None]] = None,
    triage_missing_coil: Optional[bool] = False,
) -> pathlib.Path:
    """

    This function builds an MTH5 file from FDSN client. The input dataset is described by fdsn_dataset.

    Parameters
    ----------
    fdsn_dataset: aurora.sandbox.io_helpers.fdsn_dataset.FDSNDataset
        Description of the dataset to create
    target_folder: Optional[pathlib.Path]
        The folder to create the dataset (mth5 file)
    run_id: str
        Label for the run
    force_align_streams: bool
        If True, the streams will be aligned if they are offset
    triage_units: list or None
        elements of the list should be in ["V/m to mV/km", "T to nT" ]
        These values in the list will result in an additional filter being added to
        the electric or magnetic field channels.
    triage_missing_coil: bool

    Returns
    -------
    h5_path: pathlib.Path
        The path to the mth5 that was built.
    """

    # Get Experiement
    try:
        inventory = fdsn_dataset.get_inventory(ensure_inventory_stages_are_named=True)
    # if inventory is None:
    #     print("Inventory Access Failed - NCEDC may be down")
    #     raise TypeError("None returned instead of Inventory")
    #     return None

    except Exception as e:  # FDSNException:
        logger.error(f"Exception {e}")
        # raise ValueError("NCEDC is Down, cannot build data")
        return
    translator = XMLInventoryMTExperiment()
    experiment = translator.xml_to_mt(inventory_object=inventory)

    # TRIAGE ONE-OFF ISSUE WITH UNITS
    if triage_units is not None:
        if "V/m to mV/km" in triage_units:
            experiment = triage_mt_units_electric_field(experiment)
        if "T to nT" in triage_units:
            experiment = triage_mt_units_magnetic_field(experiment)
    if triage_missing_coil:
        experiment = triage_missing_coil_hollister(experiment)

    # INITIALIZE MTH5
    h5_path = target_folder.joinpath(fdsn_dataset.h5_filebase)
    mth5_obj = initialize_mth5(h5_path)
    mth5_obj.from_experiment(experiment)

    fdsn_dataset.describe()

    # GET DATA STREAMS
    streams = fdsn_dataset.get_data_via_fdsn_client()
    streams = make_channel_labels_fdsn_compliant(streams)
    if force_align_streams:
        logger.warning("WARNING: ALIGN STREAMS NOT ROBUSTLY TESTED")
        streams = align_streams(streams, fdsn_dataset.starttime)
    streams = trim_streams_to_common_timestamps(streams)

    streams_dict = {}
    station_groups = {}
    # Iterate over stations, packing runs with data (not robust)
    for i_station, station_id in enumerate(mth5_obj.station_list):
        station_traces = [tr for tr in streams.traces if tr.stats.station == station_id]
        streams_dict[station_id] = obspy.core.Stream(station_traces)
        station_groups[station_id] = mth5_obj.get_station(station_id)
        run_metadata = experiment.surveys[0].stations[i_station].runs[0]
        run_metadata.id = (
            run_id  #  This seems to get ignored by the call to from_obspy_stream below
        )
        run_ts_obj = RunTS()
        run_ts_obj.from_obspy_stream(streams_dict[station_id], run_metadata)
        run_ts_obj.run_metadata.id = run_id  # Force setting run id
        run_group = station_groups[station_id].add_run(run_id)
        run_group.from_runts(run_ts_obj)
    mth5_obj.close_mth5()
    return h5_path



"""
    Collection of modules used in processing pipeline that operate on time series
"""
import mt_metadata
import numpy as np
import pandas as pd
import scipy.signal as ssig
import xarray as xr

from loguru import logger
from aurora.time_series.windowing_scheme import window_scheme_from_decimation
from mt_metadata.processing import TimeSeriesDecimation
from mt_metadata.processing.aurora.decimation_level import (
    DecimationLevel as AuroraDecimationLevel,
)
from mt_metadata.processing.fourier_coefficients import Decimation as FCDecimation

from mth5.groups import RunGroup
from typing import Union


def truncate_to_clock_zero(
    decimation_obj: Union[AuroraDecimationLevel, FCDecimation],
    run_xrds: RunGroup,
):
    """
    Compute the time interval between the first data sample and the clock zero
    Identify the first sample in the xarray time series that corresponds to a
    window start sample.

    Parameters
    ----------
    decimation_obj: Union[AuroraDecimationLevel, FCDecimation]
        Information about how the decimation level is to be processed
    run_xrds : xarray.core.dataset.Dataset
        normally extracted from mth5.RunTS


    Returns
    -------
    run_xrds : xarray.core.dataset.Dataset
        same as the input time series, but possibly slightly shortened
    """
    if decimation_obj.stft.window.clock_zero_type == "ignore":
        pass
    else:
        clock_zero = pd.Timestamp(decimation_obj.stft.window.clock_zero)
        clock_zero = clock_zero.to_datetime64()
        delta_t = clock_zero - run_xrds.time[0]
        assert delta_t.dtype == "<m8[ns]"  # expected in nanoseconds
        delta_t_seconds = int(delta_t) / 1e9
        if delta_t_seconds == 0:
            pass  # time series start is already clock zero
        else:
            windowing_scheme = window_scheme_from_decimation(decimation_obj)
            number_of_steps = delta_t_seconds / windowing_scheme.duration_advance
            n_partial_steps = number_of_steps - np.floor(number_of_steps)
            n_clip = n_partial_steps * windowing_scheme.num_samples_advance
            n_clip = int(np.round(n_clip))
            t_clip = run_xrds.time[n_clip]
            cond1 = run_xrds.time >= t_clip
            msg = (
                f"dropping {n_clip} samples to agree with "
                f"{decimation_obj.stft.window.clock_zero_type} clock zero {clock_zero}"
            )
            logger.info(msg)
            run_xrds = run_xrds.where(cond1, drop=True)
    return run_xrds


def prototype_decimate(
    ts_decimation: TimeSeriesDecimation,
    run_xrds: xr.Dataset,
) -> xr.Dataset:
    """
    Basically a wrapper for scipy.signal.decimate.  Takes input timeseries (as xarray
     Dataset) and a TimeSeriesDecimation object and returns a decimated version of the
     input time series.

    TODO: Consider moving this function into time_series/decimate.py
    TODO: Consider Replacing the downsampled_time_axis with rolling mean, or somthing that takes the average value of the time, not the window start
    TODO: Compare outputs with scipy resample_poly, which also has an FIR AAF and appears faster
    TODO: Add handling for case that could occur when sliced time axis has a different length than the decimated data -- see mth5 issue #217 https://github.com/kujaku11/mth5/issues/217

    Parameters
    ----------
    ts_decimation : AuroraDecimationLevel
    run_xrds: xr.Dataset
        Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
        multiple times

    Returns
    -------
    xr_ds: xr.Dataset
        Decimated version of the input run_xrds
    """
    # downsample the time axis
    slicer = slice(None, None, int(ts_decimation.factor))  # decimation.factor
    downsampled_time_axis = run_xrds.time.data[slicer]

    # decimate the time series
    num_observations = len(downsampled_time_axis)
    channel_labels = list(run_xrds.data_vars.keys())  # run_ts.channels
    num_channels = len(channel_labels)
    new_data = np.full((num_observations, num_channels), np.nan)
    for i_ch, ch_label in enumerate(channel_labels):
        # TODO: add check here for ts_decimation.anti_alias_filter
        new_data[:, i_ch] = ssig.decimate(run_xrds[ch_label], int(ts_decimation.factor))

    xr_da = xr.DataArray(
        new_data,
        dims=["time", "channel"],
        coords={"time": downsampled_time_axis, "channel": channel_labels},
    )
    attr_dict = run_xrds.attrs
    attr_dict["sample_rate"] = ts_decimation.sample_rate
    xr_da.attrs = attr_dict
    xr_ds = xr_da.to_dataset("channel")
    return xr_ds


# Here are some other ways to downsample/decimate that were experimented with.
# TODO: Jupyter notebook tutorial showing how these differ (cf standard method).
# def prototype_decimate_2(config, run_xrds):
#     """
#     Uses the built-in xarray coarsen method.   Not clear what AAF effects are.
#     Method is fast.  Might be non-linear.  Seems to give similar performance to
#     prototype_decimate for synthetic data.
#
#     N.B. config.decimation.factor must be integer valued
#
#     Parameters
#     ----------
#     config : mt_metadata.processing.aurora.Decimation
#     run_xrds: xr.Dataset
#         Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
#         multiple times
#
#     Returns
#     -------
#     xr_ds: xr.Dataset
#         Decimated version of the input run_xrds
#     """
#     new_xr_ds = run_xrds.coarsen(time=int(config.decimation.factor), boundary="trim").mean()
#     attr_dict = run_xrds.attrs
#     attr_dict["sample_rate"] = config.sample_rate
#     new_xr_ds.attrs = attr_dict
#     return new_xr_ds
#
#
# def prototype_decimate_3(config, run_xrds):
#     """
#     Uses scipy's resample method.   Not clear what AAF effects are.
#     Method is fast.
#
#     Parameters
#     ----------
#     config : mt_metadata.processing.aurora.Decimation
#     run_xrds: xr.Dataset
#         Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
#         multiple times
#
#     Returns
#     -------
#     xr_ds: xr.Dataset
#         Decimated version of the input run_xrds
#     """
#     dt = run_xrds.time.diff(dim="time").median().values
#     dt_new = config.decimation.factor * dt
#     dt_new = dt_new.__str__().replace("nanoseconds", "ns")
#     new_xr_ds = run_xrds.resample(time=dt_new).mean(dim="time")
#     attr_dict = run_xrds.attrs
#     attr_dict["sample_rate"] = config.sample_rate
#     new_xr_ds.attrs = attr_dict
#     return new_xr_ds



"""
    Computes transfer functions from covariance matrix of cross powers.

    Development notes:
    It is not uncommon to want to compute simple, "non-robust" transfer functions from matrices of cross-powers during
    MT processing.  The preliminary estimates can be used as "features" that are in turn used to compute data weights
    prior to robust regression.

    There will probably be faster, vectorial ways to do this in future versions of the software, however, in this
    work-in-progress module are the functions that can be used to generate TF estimates from matrices of cross-powers
    using xarray DataArrays as the cross-power containers.

    The idea here is to boil the TF estimation from cross powers down to only two equations.
    Equations 10-15 in Sims et al. 1971 describe the 6 ways to compute the SS TF.
    Vozoff (1991) distills this to one equation (42) with six choices for a tuple (A,B)
    (A,B) = (Ex, Ey), (Ex, Hx), (Ex, Hy), (Ey, Hx), (Ey, Hy), (Hx, Hy)

    Vozoff Notation:
                <Ex,A*><Hy,B*> - <Ex,B*><Hy,A*>
     Z_xx =     ------------------------------       Equation V91_42a
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>


                <Ex,A*><Hx,B*> - <Ex,B*><Hx,A*>
     Z_xy =     ------------------------------        Equation V91_42b (modified to correct typo in publication)
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>


                <Ey,A*><Hy,B*> - <Ey,B*><Hy,A*>
     Z_yx =     ------------------------------         Equation V91_42c
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>


                <Ey,A*><Hx,B*> - <Ey,B*><Hx,A*>
     Z_yy =     ------------------------------          Equation V91_42d
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    Note that the Equations involving Ex (Z_xx, & Zxy) are identical to the Equations
    involving Ey (Z_yx, & Zyy) if we swap Ex for Ey.  I.e.
    - Equation V91_42a is the same as Equation V91_42c if we switch Ex for Ey.  Similarly,
    - Equation V91_42b is the same as Equation V91_42d if we switch Ex for Ey.
    - This motivates a similar formulation for the Tipper:


                <Hz,A*><Hy,B*> - <Hz,B*><Hy,A*>
     T_x =     ------------------------------
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>

                <Hz,A*><Hx,B*> - <Hz,B*><Hx,A*>
     T_y =     ------------------------------
                <Hy,A*><Hx,B*> - <Hy,B*><Hy,A*>

    However, even V91 can be further distilled to only two equations, with an argument defining the field component
    that is the "output" of the TF formulation Y = one of ("Ex", "Ey", "Hz")
    I.e. by the above logic, there is one more abstraction -- instead of six functions for TF, let Ex, Ey, Hz be called say, Y.

    References:
    Sims, W.E., F.X. Bostick, and H.W. SMITH. “THE ESTIMATION OF MAGNETOTELLURIC IMPEDANCE TENSOR ELEMENTS FROM MEASURED DATA” GEOPHYSICS 36, no. 5 (1971). https://doi.org/DOI:10.1190/1.1440225.
    @inbook{doi:10.1190/1.9781560802686.ch8,
    author = {K. Vozoff},
    title = {8. The Magnetotelluric Method},
    booktitle = {Electromagnetic Methods in Applied Geophysics: Volume 2, Application, Parts A and B},
    chapter = {},
    pages = {641-712},
    year = {2012},
    doi = {10.1190/1.9781560802686.ch8},
    URL = {https://library.seg.org/doi/abs/10.1190/1.9781560802686.ch8},
    eprint = {https://library.seg.org/doi/pdf/10.1190/1.9781560802686.ch8}
    }

    TODO: Consider making the _t__x, _t__y methods of a covariance matrix class.

    TODO: Expand this to a jupyter tutorial
     -  Add Sim's Single station estimators
     - Add more doc -- For any of the TF equations, two of 6 (A,B) combinations are unstable, two are upward biased, and two are downward biased.
     - Verify that setting (A=Ey, B=Ex) recovers Sims Equation 10
     - Verify that Sims' Equations 11-15 result from an appropriate choice of (A,B)

"""

import xarray as xr
from mt_metadata.transfer_functions.tf.transfer_function import TransferFunction
from typing import Optional, Union


def tf_from_cross_powers(
    sdm: xr.DataArray,
    station_id: str,
    remote: Optional[str] = "",
    #    style: Optional[str] = "",
    components: Optional[str] = "full",
    period: Optional[float] = -1.0,
    output_format: Optional[str] = "dict",
    join_char: Optional[str] = "_",
) -> Union[dict, TransferFunction]:
    """
    returns the requested tf components from the SDM

    Develpment notes:

    TODO add validator to ensure impedance components are present and tipper components are present

    TODO: Careful here with conjugates -- we need to establish a conjugation convention for the SDM.
     Let Xi be the data from ith row of X, and Xj be the data from the jth row of X.
     Then <Xi, Xj*> will be S[i,j]
     and <Xi*, Xj> will be S[j,i]
     Vozoff only uses the form where the second variable is conjugate, so we should be consistent throughout.
     The SDM is assumed to be formed from X@XH, so S[i,j] = Xi@XjH

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    station_id: str
        The label of the station to get a TF for
    remote: str
        The label of the remote reference station to get a TF for.
    style: str
        There are six ways to estimate the single station TF
        Supported styles will be: SS[1,2,3,4,5,6], RRH, RRE, MV
        SS1-6 will use the equations from Sims (which can be simplified to fit Vozoff 1991)
        RRH will use the classic RR method
            (which simply replaces
    components: Optional[str] = "full",
        The transfer function components to return, this could be impedance, tipper, or full.
    period: float
        Not yet supported -- needed to convert tf to apparent resistivity
    join_char: str
        The character that links the station_id with the component name in the covariance matrix

    Returns
    -------

    """
    Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(station_id, remote, join_char)

    tf = {}
    tf["z_xx"] = _tf__x(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
    tf["z_xy"] = _tf__y(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
    tf["z_yx"] = _tf__x(sdm, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
    tf["z_yy"] = _tf__y(sdm, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
    tf["t_zx"] = _tf__x(sdm, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B)
    tf["t_zy"] = _tf__y(sdm, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B)

    # TODO: Add this to a test module or ipynb.
    assert _zxx(sdm, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
        sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B
    )
    assert _zxy(sdm, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
        sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B
    )
    assert _zyx(sdm, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
        sdm, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B
    )
    assert _zyy(sdm, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
        sdm, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B
    )
    assert _tx(sdm, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
        sdm, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B
    )
    assert _ty(sdm, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
        sdm, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B
    )

    if output_format == "dict":
        return tf
    elif output_format == "mt_metadata":
        # To make an mt_metadata tf, we need to pass period
        msg = "mt_metadata format is coming soon -- work in progress"
        raise NotImplementedError(msg)

        # mtm_tf = TF(period=period)
        # output = {}


def _channel_names(
    station_id: str,
    remote: Optional[str] = "",
    join_char: Optional[str] = "_",
) -> tuple:
    """

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    station_id: str
        The label of the station to get a TF for
    remote: str
        The label of the remote reference station to get a TF for.
    join_char: str
        The character that links the station_id with the component name in the covariance matrix

    Returns
    -------
    tuple: Ex, Ey, Hx, Hy, Hz, A, B the labels associated with the covariance matrix rows and columns
    """
    if remote:
        A = join_char.join((remote, "hx"))
        B = join_char.join((remote, "hy"))
    else:
        A = join_char.join((station_id, "hx"))
        B = join_char.join((station_id, "hy"))

    Ex = join_char.join((station_id, "ex"))
    Ey = join_char.join((station_id, "ey"))
    Hx = join_char.join((station_id, "hx"))
    Hy = join_char.join((station_id, "hy"))
    Hz = join_char.join((station_id, "hz"))
    return Ex, Ey, Hx, Hy, Hz, A, B


def _tf__x(sdm: xr.DataArray, Y: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """

    Compute the transfer function tensor element associated with Z_xx, Z_yx, or T_x:


                 <Y,A*><Hx,B*> -  <Y,B*><Hx,A*>
     tf__x =    ------------------------------            Generalized form of Vozoff, 1991, equation 42a
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Y: str
        The label for the "output" channel at the station in the regression formula, as referenced in the sdm.
        This label should be associated with one of the station channels ["ex", "ey", or "hz"].
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_xx, Z_yx, or T_x

    """
    numerator = sdm.loc[Y, A] * sdm.loc[Hy, B] - sdm.loc[Y, B] * sdm.loc[Hy, A]
    denominator = sdm.loc[Hx, A] * sdm.loc[Hy, B] - sdm.loc[Hx, B] * sdm.loc[Hy, A]
    return numerator / denominator


def _tf__y(sdm: xr.DataArray, Y: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """
    Compute the transfer function tensor element associated with Z_xy, Z_yy, or T_y:

                 <Y,A*><Hx,B*> - <Y,B*><Hx,A*>
     tf__y =    ------------------------------        Generalized form of Vozoff, 1991, equation 42b
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Y: str
        The label for the "output" channel at the station in the regression formula, as referenced in the sdm.
        This label should be associated with one of the station channels ["ex", "ey", or "hz"].
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_xy, Z_yy, or T_y:
    """
    numerator = sdm.loc[Y, A] * sdm.loc[Hx, B] - sdm.loc[Y, B] * sdm.loc[Hx, A]
    denominator = sdm.loc[Hy, A] * sdm.loc[Hx, B] - sdm.loc[Hy, B] * sdm.loc[Hx, A]
    return numerator / denominator


# =============================================================================
# test support methods
# =============================================================================
# TODO consider adding these to test_utils


def _zxx(sdm: xr.DataArray, Ex: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """

    Compute the transfer function tensor element Z_xx given by:

                <Ex,A*><Hy,B*> - <Ex,B*><Hy,A*>
     Z_xx =     ------------------------------       Equation V91_42a
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_xx
    """
    numerator = sdm.loc[Ex, A] * sdm.loc[Hy, B] - sdm.loc[Ex, B] * sdm.loc[Hy, A]
    denominator = sdm.loc[Hx, A] * sdm.loc[Hy, B] - sdm.loc[Hx, B] * sdm.loc[Hy, A]
    return numerator / denominator


def _zxy(sdm: xr.DataArray, Ex: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """
    Compute the transfer function tensor element Z_xy given by:

                <Ex,A*><Hx,B*> - <Ex,B*><Hx,A*>
     Z_xy =     ------------------------------        Equation V91_42b
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    **Note**: There is a typo in Vozoff 1991, Equation 42b.
    In the denominator, the final term <Hx,A*> is published as <Hy,A*>.

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_xy

    """
    numerator = sdm.loc[Ex, A] * sdm.loc[Hx, B] - sdm.loc[Ex, B] * sdm.loc[Hx, A]
    denominator = sdm.loc[Hy, A] * sdm.loc[Hx, B] - sdm.loc[Hy, B] * sdm.loc[Hx, A]
    return numerator / denominator


def _zyx(sdm: xr.DataArray, Ey: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """

    Compute the transfer function tensor element Z_yx given by:

                <Ey,A*><Hy,B*> - <Ey,B*><Hy,A*>
     Z_yx =     ------------------------------         Equation V91_42c
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_yx

    """
    numerator = sdm.loc[Ey, A] * sdm.loc[Hy, B] - sdm.loc[Ey, B] * sdm.loc[Hy, A]
    denominator = sdm.loc[Hx, A] * sdm.loc[Hy, B] - sdm.loc[Hx, B] * sdm.loc[Hy, A]
    return numerator / denominator


def _zyy(sdm: xr.DataArray, Ey: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """
    Compute the transfer function tensor element Z_yy given by:

                <Ey,A*><Hx,B*> - <Ey,B*><Hx,A*>
     Z_yy =     ------------------------------          Equation V91_42d
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element Z_yy

    """
    numerator = sdm.loc[Ey, A] * sdm.loc[Hx, B] - sdm.loc[Ey, B] * sdm.loc[Hx, A]
    denominator = sdm.loc[Hy, A] * sdm.loc[Hx, B] - sdm.loc[Hy, B] * sdm.loc[Hx, A]
    return numerator / denominator


def _tx(sdm: xr.DataArray, Hz: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """
    Compute the transfer function tensor element T_x given by:

                <Hz,A*><Hy,B*> - <Hz,B*><Hy,A*>
     T_x  =     ------------------------------
                <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element T_x

    """
    numerator = sdm.loc[Hz, A] * sdm.loc[Hy, B] - sdm.loc[Hz, B] * sdm.loc[Hy, A]
    denominator = sdm.loc[Hx, A] * sdm.loc[Hy, B] - sdm.loc[Hx, B] * sdm.loc[Hy, A]
    return numerator / denominator


def _ty(sdm: xr.DataArray, Hz: str, Hx: str, Hy: str, A: str, B: str) -> complex:
    """
    Compute the transfer function tensor element T_y given by:

                <Hz,A*><Hx,B*> - <Hz,B*><Hx,A*>
     T_y =       ------------------------------
                <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>

    Parameters
    ----------
    sdm: xr.DataArray
        A covariance matrix of the Fourier coefficients for a particular frequency band.
    Ex: str
        The label for accessing the ex data of the station associated with the TF element from the covariance matrix.
    Hx: str
        The label for accessing the hx data of the station associated with the TF element from the covariance matrix.
    Hy: str
        The label for accessing the hy data of the station associated with the TF element from the covariance matrix.
    A: str
        The label for accessing the first reference channel data.  For single station this is normally hx, for remote reference this is normally hx at the reference station.
    B: str
        The label for accessing the first reference channel data.  For single station this is normally hy, for remote reference this is normally hy at the reference station.

    Returns
    -------
    complex: transfer function tensor element T_y

    """
    numerator = sdm.loc[Hz, A] * sdm.loc[Hx, B] - sdm.loc[Hz, B] * sdm.loc[Hx, A]
    denominator = sdm.loc[Hy, A] * sdm.loc[Hx, B] - sdm.loc[Hy, B] * sdm.loc[Hx, A]
    return numerator / denominator



"""
    This module contains a class that can be used to create mth5 datasets from FDSN servers

    TODO: Review usages and consider replace with MTH5 FDSN class.
"""

from aurora.sandbox.io_helpers.inventory_review import describe_inventory_stages
from aurora.sandbox.io_helpers.inventory_review import scan_inventory_for_nonconformity
from obspy.clients.fdsn import Client
from loguru import logger


class FDSNDataset(object):
    """
    This class contains the information needed to uniquely specify a
    dataset that will be accessed from IRIS, NCEDC, or other FDSN client.
    This config will only work for single stations.

    Need:
    -fdsn_metadata_parameters
    -data_parameters (how to rover, or load from local)
    -a way to specify station-channel, this config will only work for single stations.

    """

    def __init__(self):
        """
        Constructor
        """
        self.network = None
        self.station = None
        self.channel_codes = None
        self.starttime = None
        self.endtime = None

        self.description = None
        self.dataset_id = None
        self.components_list = None  #
        self.data_source = "IRIS"

        self._client = None

    @property
    def client(self):
        """Returns the client (creates if doesn't exist)"""
        if self._client is None:
            self.initialize_client()
        return self._client

    def initialize_client(self):
        """Initialize the client"""
        self._client = Client(base_url=self.data_source, force_redirect=True)

    def get_inventory(self, ensure_inventory_stages_are_named=True, level="response"):
        """Gets inventory object from FDSN client"""
        inventory = self.client.get_stations(
            network=self.network,
            station=self.station,
            channel=self.channel_codes,
            starttime=self.starttime,
            endtime=self.endtime,
            level=level,
        )
        inventory = scan_inventory_for_nonconformity(inventory)
        if ensure_inventory_stages_are_named:
            describe_inventory_stages(inventory, assign_names=True)
        return inventory

    # def get_data_via_rover(self):
    #     """ """
    #     raise NotImplementedError

    def get_data_via_fdsn_client(self):
        """Gets data from client"""
        streams = self.client.get_waveforms(
            self.network,
            self.station,
            None,
            self.channel_codes,
            self.starttime,
            self.endtime,
        )
        return streams

    def describe(self):
        """logs some info to terminal"""
        logger.info(f"station_id = {self.station}")
        logger.info(f"network_id = {self.network}")
        logger.info(f"channel_ids = {self.channel_codes}")

    @property
    def h5_filebase(self):
        """retruns an h5 filename"""
        filebase = f"{self.dataset_id}.h5"
        return filebase



"""
    This module contains helper functions to control where the parkfield test data
    and test results are stored /accessed.

    Development Notes
    -----------------
    - Initially, the parkfield data was stored in DATA_PATH/parkfield, but this
      caused issues with write permissions on some systems (e.g., GitHub Actions runners)
      and GADI HPC systems. Therefore, the base path was changed to ~/.cache/aurora/parkfield
      to ensure that the user has write permissions.
"""

from aurora.general_helper_functions import DATA_PATH


# import pathlib


def make_parkfield_paths() -> dict:
    """
    Makes a dictionary with information about where to store/access PKD test data and results.

    Returns
    -------
    parkfield_paths: dict
        Dict containing paths to "data", "aurora_results", "config", "emtf_results"
    """
    base_path = DATA_PATH.joinpath("parkfield")
    # base_path = pathlib.Path.home().joinpath(".cache", "aurora", "parkfield")

    parkfield_paths = {}
    parkfield_paths["data"] = base_path
    parkfield_paths["aurora_results"] = base_path.joinpath("aurora_results")
    parkfield_paths["config"] = base_path.joinpath("config")
    parkfield_paths["emtf_results"] = base_path.joinpath("emtf_results")
    return parkfield_paths


PARKFIELD_PATHS = make_parkfield_paths()



"""Pytest translation of test_multi_run.py

Tests multi-run processing scenarios including individual runs, combined runs,
and runs with truncated data.
"""

import logging

import pandas as pd
import pytest
from mth5.helpers import close_open_files
from mth5.processing import KernelDataset, RunSummary

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5


@pytest.fixture(autouse=True)
def setup_logging():
    """Disable noisy matplotlib loggers."""
    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True


@pytest.fixture(scope="module")
def run_summary_test3(worker_safe_test3_h5):
    """Create a RunSummary from test3.h5 MTH5 file."""
    close_open_files()
    mth5_paths = [worker_safe_test3_h5]
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    return run_summary


class TestMultiRunProcessing:
    """Tests for multi-run processing scenarios - cache expensive process_mth5 calls."""

    @pytest.fixture(scope="class")
    def kernel_dataset_test3(self, run_summary_test3):
        """Create kernel dataset for test3."""
        kernel_dataset = KernelDataset()
        kernel_dataset.from_run_summary(run_summary_test3, "test3")
        return kernel_dataset

    @pytest.fixture(scope="class")
    def config_test3(self, kernel_dataset_test3):
        """Create config for test3 with RME estimator."""
        cc = ConfigCreator()
        return cc.create_from_kernel_dataset(
            kernel_dataset_test3, estimator={"engine": "RME"}
        )

    @pytest.fixture(scope="class")
    def processed_tf_all_runs(
        self, kernel_dataset_test3, config_test3, synthetic_test_paths
    ):
        """Process all runs together - expensive operation, done once."""
        close_open_files()
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath("syn3_all.zss")
        return process_mth5(
            config_test3,
            kernel_dataset_test3,
            units="MT",
            show_plot=False,
            z_file_path=z_file_path,
        )

    def test_all_runs(self, processed_tf_all_runs, synthetic_test_paths):
        """Test processing all runs together."""
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            "syn3_all.xml"
        )
        processed_tf_all_runs.write(fn=xml_file_name, file_type="emtfxml")


def test_each_run_individually(run_summary_test3, synthetic_test_paths, subtests):
    """Test processing each run individually.

    Note: This test must process each run separately, so it cannot use class fixtures.
    It processes 4 runs individually which is inherently expensive.
    """
    close_open_files()

    for run_id in run_summary_test3.df.run.unique():
        with subtests.test(run=run_id):
            kernel_dataset = KernelDataset()
            kernel_dataset.from_run_summary(run_summary_test3, "test3")
            station_runs_dict = {}
            station_runs_dict["test3"] = [run_id]
            keep_or_drop = "keep"
            kernel_dataset.select_station_runs(station_runs_dict, keep_or_drop)

            cc = ConfigCreator()
            config = cc.create_from_kernel_dataset(kernel_dataset)

            for decimation in config.decimations:
                decimation.estimator.engine = "RME"

            show_plot = False
            z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
                f"syn3_{run_id}.zss"
            )
            tf_cls = process_mth5(
                config,
                kernel_dataset,
                units="MT",
                show_plot=show_plot,
                z_file_path=z_file_path,
            )

            xml_file_base = f"syn3_{run_id}.xml"
            xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
                xml_file_base
            )
            tf_cls.write(fn=xml_file_name, file_type="emtfxml")


def test_works_with_truncated_run(run_summary_test3, synthetic_test_paths):
    """Test processing with a truncated run.

    Synthetic runs are 40000s long. By truncating one of the runs to 10000s,
    we make the 4th decimation invalid for that run. By truncating to 2000s
    long we make the 3rd and 4th decimation levels invalid.

    Note: This test modifies run_summary, so it cannot use class fixtures.
    """
    # Make a copy of the run summary to avoid modifying the fixture
    import copy

    run_summary = copy.deepcopy(run_summary_test3)

    delta = pd.Timedelta(seconds=38000)
    run_summary.df.loc[1, "end"] -= delta

    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, "test3")

    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(kernel_dataset, estimator={"engine": "RME"})

    show_plot = False
    z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
        "syn3_all_truncated_run.zss"
    )
    tf_cls = process_mth5(
        config,
        kernel_dataset,
        units="MT",
        show_plot=show_plot,
        z_file_path=z_file_path,
    )

    # process_mth5 may return None if insufficient data after truncation
    if tf_cls is not None:
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            "syn3_all_truncated_run.xml"
        )
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")



"""
Pytest suite for MATLAB Z-file reader functionality.

Tests reading and parsing MATLAB Z-files for different case IDs.
"""

import pytest

from aurora.sandbox.io_helpers.garys_matlab_zfiles.matlab_z_file_reader import (
    test_matlab_zfile_reader,
)


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture(params=["IAK34ss", "synthetic"])
def case_id(request):
    """
    Provide case IDs for MATLAB Z-file reader tests.

    Parameters:
    - IAK34ss: Real data case
    - synthetic: Synthetic data case
    """
    return request.param


@pytest.fixture
def iak34ss_case_id():
    """Fixture for IAK34ss case ID (real data)."""
    return "IAK34ss"


@pytest.fixture
def synthetic_case_id():
    """Fixture for synthetic case ID."""
    return "synthetic"


# =============================================================================
# Tests
# =============================================================================


def test_matlab_zfile_reader_iak34ss(iak34ss_case_id):
    """Test MATLAB Z-file reader with IAK34ss real data case."""
    test_matlab_zfile_reader(case_id=iak34ss_case_id)


@pytest.mark.skip(reason="Synthetic case currently disabled in original test")
def test_matlab_zfile_reader_synthetic(synthetic_case_id):
    """Test MATLAB Z-file reader with synthetic data case."""
    test_matlab_zfile_reader(case_id=synthetic_case_id)


@pytest.mark.parametrize("test_case_id", ["IAK34ss"])
def test_matlab_zfile_reader_parametrized(test_case_id):
    """
    Parametrized test for MATLAB Z-file reader.

    This test runs for each case ID in the parametrize decorator.
    To enable synthetic test, add "synthetic" to the parametrize list.
    """
    test_matlab_zfile_reader(case_id=test_case_id)


class TestMatlabZFileReader:
    """Test class for MATLAB Z-file reader functionality."""

    def test_iak34ss_case(self):
        """Test reading IAK34ss MATLAB Z-file."""
        test_matlab_zfile_reader(case_id="IAK34ss")

    @pytest.mark.skip(reason="Synthetic case needs verification")
    def test_synthetic_case(self):
        """Test reading synthetic MATLAB Z-file."""
        test_matlab_zfile_reader(case_id="synthetic")


# =============================================================================
# Integration Tests
# =============================================================================


class TestMatlabZFileReaderIntegration:
    """Integration tests for MATLAB Z-file reader."""

    @pytest.mark.parametrize(
        "case_id,description",
        [
            ("IAK34ss", "Real data from IAK34ss station"),
            # ("synthetic", "Synthetic test data"),  # Uncomment to enable
        ],
        ids=["IAK34ss"],  # Add "synthetic" when uncommenting above
    )
    def test_reader_with_description(self, case_id, description):
        """
        Test MATLAB Z-file reader with case descriptions.

        Parameters
        ----------
        case_id : str
            The case identifier for the MATLAB Z-file
        description : str
            Human-readable description of the test case
        """
        # Log the test case being run
        print(f"\nTesting case: {case_id} - {description}")
        test_matlab_zfile_reader(case_id=case_id)


# =============================================================================
# Backward Compatibility
# =============================================================================


def test():
    """
    Legacy test function for backward compatibility.

    This maintains the original test interface from test_matlab_zfile_reader.py
    """
    test_matlab_zfile_reader(case_id="IAK34ss")


if __name__ == "__main__":
    # Run pytest on this file
    pytest.main([__file__, "-v"])



"""
This module contains a class that helps manage data paths for testing aurora on synthetic data.

Development Notes:
 - This class was built to handle Issue #303 (installation on read-only file system). https://github.com/simpeg/aurora/issues/303

"""

from aurora.general_helper_functions import DATA_PATH
from loguru import logger
from mth5.data.paths import SyntheticTestPaths as MTH5SyntheticDataPaths

import mth5
import pathlib

DEFAULT_SANDBOX_PATH = DATA_PATH.joinpath("synthetic")


class SyntheticTestPaths(MTH5SyntheticDataPaths):
    """
    sandbox path must be a place that has write access.  This class was created because on some
    installations we only have read access.  Originally there was a data/ folder with the synthetic
    ascii data stored there, and we created the mth5 and other data products in the same place.

    Here we have a ascii_data_path which points at the ascii files (which may be read only), but we
    also accept a kwarg for "sandbox_path" which is writable and this is where the mth5 and etc. will
    get built.

    TODO: consider creating a symlink in aurora's legacy data path that points at the mth5 ascii files.

    """

    def __init__(self, sandbox_path=DEFAULT_SANDBOX_PATH, ascii_data_path=None):
        """
        Constructor

        Parameters
        ----------
        sandbox_path: pathlib.Path
            A writable path where test results are stored

        ivars:
        - ascii_data_path: where the ascii data are stored
        - mth5_path: this is where the mth5 files get written to.
        - config_path: this is where the config files get saved while tests are running
        - aurora_results_path: this is where the processing results get saved during test execution
        - emtf_results_path: stores some legacy results from EMTF processing for tests/comparison.

        """
        super(
            MTH5SyntheticDataPaths, self
        ).__init__()  # sandbox_path=sandbox_path,ascii_data_path=ascii_data_path)
        # READ ONLY OK
        if ascii_data_path is None:
            self.ascii_data_path = _get_mth5_ascii_data_path()

        # NEED WRITE ACCESS
        # Consider using an environment variable for sandbox_path
        if sandbox_path is None:
            logger.debug(
                f"synthetic sandbox path is being set to {DEFAULT_SANDBOX_PATH}"
            )
            self._sandbox_path = DEFAULT_SANDBOX_PATH
        else:
            self._sandbox_path = sandbox_path

        self.mth5_path = self._sandbox_path.joinpath("mth5")
        self.aurora_results_path = self._sandbox_path.joinpath("aurora_results")
        self.emtf_results_path = self._sandbox_path.joinpath("emtf_results")
        self.config_path = self._sandbox_path.joinpath("config")
        self.writability_check()
        # assert self.ascii_data_path.exists()

    def writability_check(self):
        """
        Placeholder
        Should check that sandbox and dirs below have write access
        If dirs are not writeable, consider
        HOME = pathlib.Path().home()
        workaround_sandbox = HOME.joinpath(".cache", "aurora", "sandbox")
        """
        pass

    def mkdirs(self):
        """
        Makes the directories that the tests will write results to.

        """
        self.aurora_results_path.mkdir(parents=True, exist_ok=True)
        self.config_path.mkdir(parents=True, exist_ok=True)
        self.emtf_results_path.mkdir(parents=True, exist_ok=True)
        self.mth5_path.mkdir(parents=True, exist_ok=True)


def _get_mth5_ascii_data_path() -> pathlib.Path:
    """
    Get the path to the ascii synthetic data

    Returns
    -------
    mth5_data_path: pathlib.Path
        This is the place where the legacy test files (ascii MT data from EMTF) are archived
    """
    mth5_data_path = pathlib.Path(mth5.__file__).parent.joinpath("data")
    return mth5_data_path



# -*- coding: utf-8 -*-
"""
Pytest suite for regression helper_functions module.

Tests transfer function regression methods using fixtures and subtests.
Optimized for pytest-xdist parallel execution.
"""

import numpy as np
import pytest

from aurora.transfer_function.regression.helper_functions import (
    direct_solve_tf,
    rme_beta,
    simple_solve_tf,
)


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture(scope="module")
def sample_electric_data():
    """Sample electric field data for testing."""
    return np.array(
        [
            4.39080123e-07 - 2.41097397e-06j,
            -2.33418464e-06 + 2.10752581e-06j,
            1.38642624e-06 - 1.87333571e-06j,
        ]
    )


@pytest.fixture(scope="module")
def sample_magnetic_data():
    """Sample magnetic field data for testing."""
    return np.array(
        [
            [7.00767250e-07 - 9.18819198e-07j, 1.94321684e-07 + 3.71934877e-07j],
            [-1.06648904e-07 + 8.19420154e-07j, 1.15361101e-08 - 6.32581646e-07j],
            [-1.02700963e-07 - 3.73904463e-07j, 3.86095787e-08 + 4.33155345e-07j],
        ]
    )


@pytest.fixture(scope="module")
def expected_solution():
    """Expected transfer function solution for sample data."""
    return np.array([-0.04192569 - 0.36502722j, -3.65284496 - 4.05194938j])


@pytest.fixture(scope="module")
def simple_2x2_system():
    """Simple 2x2 system for basic testing."""
    X = np.array([[1.0 + 0j, 0.0 + 0j], [0.0 + 0j, 1.0 + 0j]])
    Y = np.array([2.0 + 1j, 3.0 - 2j])
    expected = Y.copy()
    return X, Y, expected


@pytest.fixture(scope="module")
def overdetermined_system():
    """Overdetermined system (more equations than unknowns)."""
    np.random.seed(42)
    X = np.random.randn(10, 2) + 1j * np.random.randn(10, 2)
    true_tf = np.array([1.5 + 0.5j, -0.8 + 1.2j])
    Y = X @ true_tf
    return X, Y, true_tf


@pytest.fixture(scope="module")
def remote_reference_data():
    """Data with remote reference channels."""
    np.random.seed(43)
    X = np.random.randn(5, 2) + 1j * np.random.randn(5, 2)
    R = np.random.randn(5, 2) + 1j * np.random.randn(5, 2)
    true_tf = np.array([2.0 + 0j, -1.0 + 0.5j])
    Y = X @ true_tf
    return X, Y, R, true_tf


# =============================================================================
# Test RME Beta Function
# =============================================================================


class TestRMEBeta:
    """Test the rme_beta correction factor function."""

    def test_rme_beta_standard_value(self):
        """Test rme_beta with standard r0=1.5."""
        beta = rme_beta(1.5)
        # For r0=1.5, beta should be approximately 0.78
        assert isinstance(beta, (float, np.floating))
        assert 0.75 < beta < 0.80
        # More precise check
        expected = 1.0 - np.exp(-1.5)
        assert np.isclose(beta, expected)

    def test_rme_beta_zero(self):
        """Test rme_beta with r0=0."""
        beta = rme_beta(0.0)
        # For r0=0, beta = 1 - exp(0) = 1 - 1 = 0
        assert np.isclose(beta, 0.0)

    def test_rme_beta_large_value(self):
        """Test rme_beta with large r0."""
        beta = rme_beta(10.0)
        # For large r0, beta approaches 1.0
        assert isinstance(beta, (float, np.floating))
        assert beta > 0.99
        expected = 1.0 - np.exp(-10.0)
        assert np.isclose(beta, expected)

    def test_rme_beta_small_value(self):
        """Test rme_beta with small positive r0."""
        beta = rme_beta(0.1)
        expected = 1.0 - np.exp(-0.1)
        assert np.isclose(beta, expected)
        # Small r0 should give small beta
        assert 0.0 < beta < 0.1

    def test_rme_beta_range_values(self, subtests):
        """Test rme_beta across a range of r0 values."""
        r0_values = [0.5, 1.0, 1.5, 2.0, 3.0, 5.0]

        for r0 in r0_values:
            with subtests.test(r0=r0):
                beta = rme_beta(r0)
                expected = 1.0 - np.exp(-r0)
                assert np.isclose(beta, expected)
                # Beta should always be in [0, 1)
                assert 0.0 <= beta < 1.0

    def test_rme_beta_monotonic(self):
        """Test that rme_beta is monotonically increasing."""
        r0_values = np.linspace(0, 5, 20)
        beta_values = [rme_beta(r0) for r0 in r0_values]

        # Check that each value is greater than or equal to previous
        for i in range(1, len(beta_values)):
            assert beta_values[i] >= beta_values[i - 1]

    def test_rme_beta_asymptotic_behavior(self):
        """Test that rme_beta approaches 1.0 asymptotically."""
        large_r0 = 100.0
        beta = rme_beta(large_r0)
        assert np.isclose(beta, 1.0, rtol=1e-10)


# =============================================================================
# Test Simple Solve TF
# =============================================================================


class TestSimpleSolveTF:
    """Test the simple_solve_tf function."""

    def test_simple_solve_tf_sample_data(
        self, sample_electric_data, sample_magnetic_data, expected_solution
    ):
        """Test simple_solve_tf with provided sample data."""
        z = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        assert np.allclose(z, expected_solution, rtol=1e-8)

    def test_simple_solve_tf_identity_system(self, simple_2x2_system):
        """Test simple_solve_tf with identity-like system."""
        X, Y, expected = simple_2x2_system
        z = simple_solve_tf(Y, X)
        assert np.allclose(z, expected, rtol=1e-10)

    def test_simple_solve_tf_overdetermined(self, overdetermined_system):
        """Test simple_solve_tf with overdetermined system."""
        X, Y, true_tf = overdetermined_system
        z = simple_solve_tf(Y, X)
        # Should recover the true TF exactly (no noise added)
        assert np.allclose(z, true_tf, rtol=1e-10)

    def test_simple_solve_tf_with_remote_reference(self, remote_reference_data):
        """Test simple_solve_tf with remote reference."""
        X, Y, R, true_tf = remote_reference_data
        # Using remote reference R instead of X for conjugate transpose
        z = simple_solve_tf(Y, X, R=R)

        # Result depends on R, not necessarily equal to true_tf
        assert z.shape == true_tf.shape
        assert np.all(np.isfinite(z))

    def test_simple_solve_tf_return_type(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that simple_solve_tf returns numpy array."""
        z = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        assert isinstance(z, np.ndarray)
        assert z.dtype == np.complex128 or z.dtype == np.complex64

    def test_simple_solve_tf_shape(self, sample_electric_data, sample_magnetic_data):
        """Test that simple_solve_tf returns correct shape."""
        z = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        # Should return 2 elements for 2-column input
        assert z.shape == (2,)

    def test_simple_solve_tf_no_remote_reference(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test simple_solve_tf explicitly with R=None."""
        z1 = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        z2 = simple_solve_tf(sample_electric_data, sample_magnetic_data, R=None)
        assert np.allclose(z1, z2)


# =============================================================================
# Test Direct Solve TF
# =============================================================================


class TestDirectSolveTF:
    """Test the direct_solve_tf function."""

    def test_direct_solve_tf_sample_data(
        self, sample_electric_data, sample_magnetic_data, expected_solution
    ):
        """Test direct_solve_tf with provided sample data."""
        z = direct_solve_tf(sample_electric_data, sample_magnetic_data)
        assert np.allclose(z, expected_solution, rtol=1e-8)

    def test_direct_solve_tf_identity_system(self, simple_2x2_system):
        """Test direct_solve_tf with identity-like system."""
        X, Y, expected = simple_2x2_system
        z = direct_solve_tf(Y, X)
        assert np.allclose(z, expected, rtol=1e-10)

    def test_direct_solve_tf_overdetermined(self, overdetermined_system):
        """Test direct_solve_tf with overdetermined system."""
        X, Y, true_tf = overdetermined_system
        z = direct_solve_tf(Y, X)
        # Should recover the true TF exactly (no noise added)
        assert np.allclose(z, true_tf, rtol=1e-10)

    def test_direct_solve_tf_with_remote_reference(self, remote_reference_data):
        """Test direct_solve_tf with remote reference."""
        X, Y, R, true_tf = remote_reference_data
        # Using remote reference R instead of X for conjugate transpose
        z = direct_solve_tf(Y, X, R=R)

        # Result depends on R, not necessarily equal to true_tf
        assert z.shape == true_tf.shape
        assert np.all(np.isfinite(z))

    def test_direct_solve_tf_return_type(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that direct_solve_tf returns numpy array."""
        z = direct_solve_tf(sample_electric_data, sample_magnetic_data)
        assert isinstance(z, np.ndarray)
        assert z.dtype == np.complex128 or z.dtype == np.complex64

    def test_direct_solve_tf_shape(self, sample_electric_data, sample_magnetic_data):
        """Test that direct_solve_tf returns correct shape."""
        z = direct_solve_tf(sample_electric_data, sample_magnetic_data)
        # Should return 2 elements for 2-column input
        assert z.shape == (2,)

    def test_direct_solve_tf_no_remote_reference(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test direct_solve_tf explicitly with R=None."""
        z1 = direct_solve_tf(sample_electric_data, sample_magnetic_data)
        z2 = direct_solve_tf(sample_electric_data, sample_magnetic_data, R=None)
        assert np.allclose(z1, z2)


# =============================================================================
# Test Equivalence Between Methods
# =============================================================================


class TestMethodEquivalence:
    """Test that simple_solve_tf and direct_solve_tf produce equivalent results."""

    def test_methods_equivalent_sample_data(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that both methods give same result on sample data."""
        z_simple = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        z_direct = direct_solve_tf(sample_electric_data, sample_magnetic_data)
        assert np.allclose(z_simple, z_direct, rtol=1e-10)

    def test_methods_equivalent_identity(self, simple_2x2_system):
        """Test that both methods give same result on identity system."""
        X, Y, _ = simple_2x2_system
        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)
        assert np.allclose(z_simple, z_direct, rtol=1e-10)

    def test_methods_equivalent_overdetermined(self, overdetermined_system):
        """Test that both methods give same result on overdetermined system."""
        X, Y, _ = overdetermined_system
        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)
        assert np.allclose(z_simple, z_direct, rtol=1e-10)

    def test_methods_equivalent_with_remote(self, remote_reference_data):
        """Test that both methods give same result with remote reference."""
        X, Y, R, _ = remote_reference_data
        z_simple = simple_solve_tf(Y, X, R=R)
        z_direct = direct_solve_tf(Y, X, R=R)
        assert np.allclose(z_simple, z_direct, rtol=1e-10)


# =============================================================================
# Test Edge Cases
# =============================================================================


class TestEdgeCases:
    """Test edge cases and boundary conditions."""

    def test_single_equation_system(self):
        """Test with minimum size system (1 equation, but need at least 2 for 2 unknowns)."""
        # Actually need at least 2 equations for 2 unknowns
        X = np.array([[1.0 + 0j, 2.0 + 0j], [3.0 + 0j, 4.0 + 0j]])
        Y = np.array([5.0 + 1j, 6.0 + 2j])

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        # Both should produce valid results
        assert np.all(np.isfinite(z_simple))
        assert np.all(np.isfinite(z_direct))
        assert np.allclose(z_simple, z_direct)

    def test_real_valued_inputs(self):
        """Test with real-valued (not complex) inputs."""
        X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        Y = np.array([7.0, 8.0, 9.0])

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        assert np.all(np.isfinite(z_simple))
        assert np.all(np.isfinite(z_direct))
        assert np.allclose(z_simple, z_direct)

    def test_complex_phases(self, subtests):
        """Test with various complex phase relationships."""
        phases = [0, np.pi / 4, np.pi / 2, np.pi]

        for phase in phases:
            with subtests.test(phase=phase):
                X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]) * np.exp(1j * phase)
                Y = np.array([1.0, 2.0, 3.0]) * np.exp(1j * (phase + np.pi / 6))

                z_simple = simple_solve_tf(Y, X)
                z_direct = direct_solve_tf(Y, X)

                assert np.all(np.isfinite(z_simple))
                assert np.all(np.isfinite(z_direct))
                assert np.allclose(z_simple, z_direct)

    def test_large_magnitude_values(self):
        """Test with very large magnitude values."""
        scale = 1e10
        X = np.array([[1.0 + 1j, 2.0 - 1j], [3.0 + 0j, 4.0 + 2j]]) * scale
        Y = np.array([5.0 + 1j, 6.0 - 2j]) * scale

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        assert np.all(np.isfinite(z_simple))
        assert np.all(np.isfinite(z_direct))
        assert np.allclose(z_simple, z_direct, rtol=1e-6)

    def test_small_magnitude_values(self):
        """Test with very small magnitude values."""
        scale = 1e-10
        X = np.array([[1.0 + 1j, 2.0 - 1j], [3.0 + 0j, 4.0 + 2j]]) * scale
        Y = np.array([5.0 + 1j, 6.0 - 2j]) * scale

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        assert np.all(np.isfinite(z_simple))
        assert np.all(np.isfinite(z_direct))
        assert np.allclose(z_simple, z_direct, rtol=1e-6)


# =============================================================================
# Test Numerical Stability
# =============================================================================


class TestNumericalStability:
    """Test numerical stability of the solvers."""

    def test_well_conditioned_system(self):
        """Test with a well-conditioned system."""
        np.random.seed(44)
        # Create well-conditioned matrix
        X = np.random.randn(10, 2) + 1j * np.random.randn(10, 2)
        X[:, 0] = X[:, 0] / np.linalg.norm(X[:, 0])
        X[:, 1] = X[:, 1] / np.linalg.norm(X[:, 1])

        true_tf = np.array([1.0 + 0.5j, -0.5 + 1.0j])
        Y = X @ true_tf

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        assert np.allclose(z_simple, true_tf, rtol=1e-10)
        assert np.allclose(z_direct, true_tf, rtol=1e-10)

    def test_orthogonal_columns(self):
        """Test with orthogonal column vectors."""
        # Create orthogonal columns
        X = np.array([[1.0, 0.0], [0.0, 1.0], [0.0, 0.0]], dtype=complex)
        Y = np.array([2.0 + 1j, 3.0 - 2j, 0.0])

        z_simple = simple_solve_tf(Y, X)
        z_direct = direct_solve_tf(Y, X)

        # For orthogonal X, solution should be straightforward
        assert np.allclose(z_simple, z_direct)
        assert np.allclose(z_simple[0], 2.0 + 1j)
        assert np.allclose(z_simple[1], 3.0 - 2j)

    def test_consistency_across_seeds(self, subtests):
        """Test that results are consistent across different random seeds."""
        seeds = [10, 20, 30, 40, 50]

        for seed in seeds:
            with subtests.test(seed=seed):
                np.random.seed(seed)
                X = np.random.randn(8, 2) + 1j * np.random.randn(8, 2)
                true_tf = np.array([1.0 + 1.0j, -1.0 + 1.0j])
                Y = X @ true_tf

                z_simple = simple_solve_tf(Y, X)
                z_direct = direct_solve_tf(Y, X)

                assert np.allclose(z_simple, true_tf, rtol=1e-10)
                assert np.allclose(z_direct, true_tf, rtol=1e-10)
                assert np.allclose(z_simple, z_direct)


# =============================================================================
# Test Data Integrity
# =============================================================================


class TestDataIntegrity:
    """Test that functions don't modify input data."""

    def test_simple_solve_tf_preserves_inputs(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that simple_solve_tf doesn't modify input arrays."""
        Y_orig = sample_electric_data.copy()
        X_orig = sample_magnetic_data.copy()

        simple_solve_tf(sample_electric_data, sample_magnetic_data)

        assert np.allclose(sample_electric_data, Y_orig)
        assert np.allclose(sample_magnetic_data, X_orig)

    def test_direct_solve_tf_preserves_inputs(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that direct_solve_tf doesn't modify input arrays."""
        Y_orig = sample_electric_data.copy()
        X_orig = sample_magnetic_data.copy()

        direct_solve_tf(sample_electric_data, sample_magnetic_data)

        assert np.allclose(sample_electric_data, Y_orig)
        assert np.allclose(sample_magnetic_data, X_orig)

    def test_remote_reference_preserved(self, remote_reference_data):
        """Test that remote reference array is not modified."""
        X, Y, R, _ = remote_reference_data
        R_orig = R.copy()

        simple_solve_tf(Y, X, R=R)
        direct_solve_tf(Y, X, R=R)

        assert np.allclose(R, R_orig)


# =============================================================================
# Test Mathematical Properties
# =============================================================================


class TestMathematicalProperties:
    """Test mathematical properties of the regression."""

    def test_linearity(self):
        """Test that the solution is linear in Y."""
        X = np.array([[1.0 + 0j, 2.0 + 0j], [3.0 + 0j, 4.0 + 0j]])
        Y1 = np.array([1.0 + 1j, 2.0 + 2j])
        Y2 = np.array([3.0 - 1j, 4.0 - 2j])

        z1 = simple_solve_tf(Y1, X)
        z2 = simple_solve_tf(Y2, X)
        z_sum = simple_solve_tf(Y1 + Y2, X)

        # Solution should be linear: z(Y1 + Y2) = z(Y1) + z(Y2)
        assert np.allclose(z_sum, z1 + z2, rtol=1e-10)

    def test_scaling_property(self):
        """Test that scaling Y scales the solution proportionally."""
        X = np.array([[1.0 + 0j, 2.0 + 0j], [3.0 + 0j, 4.0 + 0j]])
        Y = np.array([1.0 + 1j, 2.0 + 2j])
        scale = 5.0 + 3j

        z1 = simple_solve_tf(Y, X)
        z2 = simple_solve_tf(scale * Y, X)

        # Scaling Y should scale the solution
        assert np.allclose(z2, scale * z1, rtol=1e-10)

    def test_residual_minimization(self):
        """Test that the solution minimizes the residual in least squares sense."""
        np.random.seed(45)
        X = np.random.randn(10, 2) + 1j * np.random.randn(10, 2)
        true_tf = np.array([1.0 + 0.5j, -0.5 + 1.0j])
        Y = X @ true_tf

        z = simple_solve_tf(Y, X)
        residual = Y - X @ z

        # Residual should be very small (near zero for exact case)
        assert np.linalg.norm(residual) < 1e-10

    def test_conjugate_transpose_property(self):
        """Test the conjugate transpose operations in the formulation."""
        X = np.array([[1.0 + 1j, 2.0 - 1j], [3.0 + 0j, 4.0 + 2j]])
        Y = np.array([5.0 + 1j, 6.0 - 2j])

        # Verify that X^H @ X is Hermitian
        xH = X.conjugate().transpose()
        xHx = xH @ X

        assert np.allclose(xHx, xHx.conj().T, rtol=1e-10)


# =============================================================================
# Test Return Value Characteristics
# =============================================================================


class TestReturnValues:
    """Test characteristics of return values."""

    def test_return_value_finite(self, sample_electric_data, sample_magnetic_data):
        """Test that return values are finite."""
        z_simple = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        z_direct = direct_solve_tf(sample_electric_data, sample_magnetic_data)

        assert np.all(np.isfinite(z_simple))
        assert np.all(np.isfinite(z_direct))

    def test_return_value_complex(self, sample_electric_data, sample_magnetic_data):
        """Test that return values are complex."""
        z_simple = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        z_direct = direct_solve_tf(sample_electric_data, sample_magnetic_data)

        assert np.iscomplexobj(z_simple)
        assert np.iscomplexobj(z_direct)

    def test_return_value_not_all_zero(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that return values are not all zero."""
        z_simple = simple_solve_tf(sample_electric_data, sample_magnetic_data)
        z_direct = direct_solve_tf(sample_electric_data, sample_magnetic_data)

        assert not np.allclose(z_simple, 0)
        assert not np.allclose(z_direct, 0)


# =============================================================================
# Test Deterministic Behavior
# =============================================================================


class TestDeterministicBehavior:
    """Test that functions produce deterministic results."""

    def test_simple_solve_tf_deterministic(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that simple_solve_tf produces same result on repeated calls."""
        results = [
            simple_solve_tf(sample_electric_data, sample_magnetic_data)
            for _ in range(5)
        ]

        for result in results[1:]:
            assert np.allclose(result, results[0])

    def test_direct_solve_tf_deterministic(
        self, sample_electric_data, sample_magnetic_data
    ):
        """Test that direct_solve_tf produces same result on repeated calls."""
        results = [
            direct_solve_tf(sample_electric_data, sample_magnetic_data)
            for _ in range(5)
        ]

        for result in results[1:]:
            assert np.allclose(result, results[0])

    def test_rme_beta_deterministic(self):
        """Test that rme_beta produces same result on repeated calls."""
        r0 = 1.5
        results = [rme_beta(r0) for _ in range(10)]

        for result in results[1:]:
            assert result == results[0]



"""
This module contains from miscellaneous functions and some global paths.
"""
import inspect
import pathlib
import subprocess
from pathlib import Path

import mt_metadata
import scipy.io as sio
from loguru import logger

import aurora


init_file = inspect.getfile(aurora)
AURORA_PATH = Path(init_file).parent.parent
DATA_PATH = AURORA_PATH.joinpath("data")
DOCS_PATH = AURORA_PATH.joinpath("docs")
TEST_PATH = AURORA_PATH.joinpath("tests")
CONFIG_PATH = AURORA_PATH.joinpath("aurora", "config")
BAND_SETUP_PATH = CONFIG_PATH.joinpath("emtf_band_setup")
PROCESSING_TEMPLATES_PATH = CONFIG_PATH.joinpath("templates")

mt_metadata_init_file = inspect.getfile(mt_metadata)
MT_METADATA_PATH = pathlib.Path(mt_metadata_init_file).parent.parent
MT_METADATA_FEATURES_TEST_HELPERS_PATH = MT_METADATA_PATH.joinpath(
    "mt_metadata", "features", "test_helpers"
)


def get_test_path() -> pathlib.Path:
    """
    Gets the path to where the test are.

    Returns
    -------
    test_path: pathlib.Path
        Object that points to where aurora's tests are.
    """
    test_path = AURORA_PATH.joinpath("tests")
    if not test_path.exists():
        msg = (
            f"Could not locate test directory {TEST_PATH}\n "
            f"This is most likely because aurora was installed from pypi or conda forge\n"
            f"TEST_PATH should be replaced with DATA_PATH"
        )
        logger.warning(msg)
    return test_path


try:
    FIGURES_PATH = DATA_PATH.joinpath("figures")
    FIGURES_PATH.mkdir(exist_ok=True, parents=True)
except OSError:
    FIGURES_PATH = None
mt_metadata_init = inspect.getfile(mt_metadata)
MT_METADATA_DATA = Path(mt_metadata_init).parent.parent.joinpath("data")


def execute_subprocess(cmd, **kwargs):
    """
    A wrapper for subprocess.call

    Parameters
    ----------
    cmd : string
        command as it would be typed in a terminal
    kwargs: denotes keyword arguments that would be passed to subprocess

    """
    exit_status = subprocess.call([cmd], shell=True, **kwargs)
    if exit_status != 0:
        raise Exception("Failed to execute \n {}".format(cmd))
    return


def replace_in_file(file_path: pathlib.Path, old: str, new: str) -> None:
    """
    Replace all instances of 'old' with 'new' in the given file.
    :param file_path: Path to the file where replacements should be made.

    """
    if not file_path.exists():
        logger.warning(f"File not found: {file_path}")
        return

    with open(file_path, "r", encoding="utf-8") as f:
        lines: list[str] = f.readlines()

    updated: list[str] = [line.replace(old, new) for line in lines]

    with open(file_path, "w", encoding="utf-8") as f:
        f.writelines(updated)

    logger.info(f"Updated: {file_path}")


def save_to_mat(data, variable_name, filename):
    """
    Saves numpy array in matlab format.

    Example Usage:
    x = X.to_array(dim="channel")
    save_to_mat(x.data, "x", "x.mat")

    Reading into matlab or Octave:
    tmp = load("x.mat");
    data = tmp.x;

    Parameters
    ----------
    data : numpy array
        the data to save to file.  its fine if this is complex-valued.
    variable_name : string
        The name that we use to reference the variable within the struct in the matfile.
    filename : string
        The filepath to output

    """
    sio.savemat(filename, {variable_name: data})
    return


class DotDict(dict):
    """
    Helper function for debugging, casts a dict so that its values
    can be accessed via dict.key as well as dict["key"]

    Usage:
    dot_dict = DotDict(basic_dict)
    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name, value):
        self[name] = value



# import logging

# import filecmp
import json
import pandas as pd
import unittest

from aurora.config.config_creator import ConfigCreator
from aurora.config.config_creator import SUPPORTED_BAND_SPECIFICATION_STYLES

from aurora.config.metadata import Processing
from aurora.config.metadata.processing import _processing_obj_from_json_file
from aurora.general_helper_functions import PROCESSING_TEMPLATES_PATH
from aurora.test_utils.synthetic.processing_helpers import get_example_kernel_dataset


class TestConfigCreator(unittest.TestCase):
    """ """

    @classmethod
    def setUpClass(self):
        from aurora.config.config_creator import SUPPORTED_BAND_SPECIFICATION_STYLES

        pass

    def setUp(self):
        pass

    def test_supported_band_specification_styles(self):
        cc = ConfigCreator()
        # confirm that we are restricting scope of styles
        with self.assertRaises(NotImplementedError):
            cc.band_specification_style = "some unsupported style"
        # confirm that supported styles ok
        for supported_style in SUPPORTED_BAND_SPECIFICATION_STYLES:
            cc.band_specification_style = supported_style
            cc.determine_band_specification_style()

    def test_exception_for_non_unique_band_specification(self):
        """
        tests that bands shouldn't be defined in two different ways

        """
        from aurora.config.emtf_band_setup import BANDS_DEFAULT_FILE

        kernel_dataset = get_example_kernel_dataset()
        cc = ConfigCreator()
        cfg1 = cc.create_from_kernel_dataset(
            kernel_dataset, estimator={"engine": "RME"}
        )

        with self.assertRaises(ValueError):
            cc2 = ConfigCreator(
                emtf_band_file=BANDS_DEFAULT_FILE, band_edges=cfg1.band_edges_dict
            )
            cc2.determine_band_specification_style()

    def test_frequency_bands(self):
        """
        Tests the frequency_bands method of AuroraDecimationLevel
        TODO: Move this into mt_metadata.
         - Requires a test in mt_metadata that creates a fully populated AuroraDecimationLevel

        """

        kernel_dataset = get_example_kernel_dataset(num_stations=1)
        cc = ConfigCreator()
        cfg1 = cc.create_from_kernel_dataset(
            kernel_dataset, estimator={"engine": "RME"}
        )
        dec_level_0 = cfg1.decimations[0]
        band_edges_a = dec_level_0.frequency_bands_obj().band_edges

        # compare with another way to get band edges
        delta_f = dec_level_0.frequency_sample_interval
        lower_edges = (dec_level_0.lower_bounds * delta_f) - delta_f / 2.0
        upper_edges = (dec_level_0.upper_bounds * delta_f) + delta_f / 2.0
        band_edges_b = pd.DataFrame(
            data={
                "lower_bound": lower_edges,
                "upper_bound": upper_edges,
            }
        )
        assert (band_edges_b - band_edges_a == 0).all().all()

    def test_default_synthetic_processing_parameters(self):
        """
            Test that the config can be
        Returns
        -------

        """
        kernel_dataset = get_example_kernel_dataset(num_stations=2)
        cc = ConfigCreator()
        processing_config = cc.create_from_kernel_dataset(
            kernel_dataset,
            estimator={"engine": "RME_RR"},
        )
        target_file = PROCESSING_TEMPLATES_PATH.joinpath("tmp_processing_config.json")
        processing_config.save_as_json(target_file)

        # get mt_processing_parameters template
        reference_file = PROCESSING_TEMPLATES_PATH.joinpath(
            "processing_configuration_template.json"
        )
        assert reference_file.exists()
        reference_processing_obj = _processing_obj_from_json_file(reference_file)

        # Now we want to assert that the config creator output matches the reference file.
        # This may fail on github since the mth5_paths will be /home/runner/some_git_path whereas locally /home/kkappler
        # in that case, may need to create something to check only non-system specific things , like:
        assert reference_processing_obj.id == processing_config.id
        assert (
            reference_processing_obj.band_specification_style
            == processing_config.band_specification_style
        )
        assert (
            reference_processing_obj.channel_nomenclature
            == processing_config.channel_nomenclature
        )
        assert reference_processing_obj.decimations == processing_config.decimations
        assert reference_processing_obj.decimations == processing_config.decimations
        assert (
            reference_processing_obj.stations.local.id
            == processing_config.stations.local.id
        )
        assert (
            reference_processing_obj.stations.local.remote
            == processing_config.stations.local.remote
        )
        assert (
            reference_processing_obj.stations.local.runs
            == processing_config.stations.local.runs
        )
        assert (
            reference_processing_obj.stations.remote[0].id
            == processing_config.stations.remote[0].id
        )
        assert (
            reference_processing_obj.stations.remote[0].remote
            == processing_config.stations.remote[0].remote
        )
        assert (
            reference_processing_obj.stations.remote[0].runs
            == processing_config.stations.remote[0].runs
        )

        # keys to ignore, because they have paths from local machine:..
        # 'band_setup_file',
        # 'stations.local.mth5_path',
        # 'stations.remote[0].mth5_path (if it exists)
        # TODO: consider simply updating these paths during the test and simply:
        # assert reference_processing_obj == processing_config
        # OR
        # assert reference_processing_obj.__eq__(
        #     processing_config,
        #     ignore_keys=['band_setup_file',
        #                  'stations.local.mth5_path',
        #                  # 'stations.remote',
        #                  ])
        # but this will still have the remote reference path ... ugh:/
        target_file.unlink()


def main():
    # tmp = TestConfigCreator()
    # tmp.setUpClass()
    # tmp.test_exception_for_non_unique_band_specification()
    unittest.main()


if __name__ == "__main__":
    main()



"""
Placeholder module for methods manipulating xarray time series
"""

import numpy as np
import xarray as xr
from loguru import logger
from typing import Optional


def handle_nan(
    X: xr.Dataset,
    Y: Optional[xr.Dataset],
    RR: Optional[xr.Dataset],
    drop_dim: Optional[str] = "",
) -> tuple:
    """
    Drops Nan from multiple channel series.

    Initial use case is for Fourier coefficients, but could be more general.

    Idea is to merge X,Y,RR together, and then call dropna.  We have to be careful
    with merging because there can be namespace clashes in the channel labels.
    Currently handling this by relabelling the remote reference channels from for
    example "hx"--> "remote_hx", "hy"-->"remote_hy".  If needed we could add "local" to
    local the other channels in X, Y.

    It would be nice to maintain an index of what was dropped.

    Development Note: Even if there are no NaNs in the data, this function can still encounter issues,
    in particular if the time coordinates of X and RR do not match exactly.
    In this case, the merge will fail, and we will log an error message.

    TODO: This should be handled more gracefully, perhaps by aligning the time coordinates
    before merging, or by using a different merge strategy that allows for slight mismatches.
    Alternatively, a validator could be added to ensure that the time coordinates of X and RR match before calling this function.
    That validation could be done by tfk, or at least immediately following the construction / access of the FCs.

    TODO: Merge this with drop_nan in transfer_function_helpers.py

    Parameters
    ----------
    X : xr.Dataset
    Y : xr.Dataset or None
    RR : xr.Dataset or None
    drop_dim: string
        specifies the dimension on which dropna is happening.  For 3D STFT arrays
        this is "time", for 2D stacked STFT this is "observation"

    Returns
    -------
    X : xr.Dataset
    Y : xr.Dataset
    RR : xr.Dataset or None

    """
    time_axis_problem = False
    if not time_axis_match(X, Y, RR):
        time_axis_problem = True
        logger.error("Time axes do not match, merging datasets with left join")
        logger.error("This may lead to unexpected results, please check your data.")

    if Y is None:
        Y = xr.Dataset()
    if RR is None:
        RR = xr.Dataset()

    input_channels = list(X.data_vars)
    output_channels = list(Y.data_vars)
    reference_channels = list(RR.data_vars)
    data_var_add_label_mapper = {}
    data_var_rm_label_mapper = {}
    for ch in reference_channels:
        data_var_add_label_mapper[ch] = f"remote_{ch}"
        data_var_rm_label_mapper[f"remote_{ch}"] = ch
    RR = RR.rename(data_var_add_label_mapper)

    merged_xr = X.merge(Y, join="exact")
    if time_axis_problem:
        # If the time axes do not match, we cannot merge them directly.
        # Instead, we will merge RR with a left join, which will keep all timestamps from
        # X and Y, and fill in the missing values from RR where available.
        merged_xr = merged_xr.merge(RR, join="left")
        for ch in list(RR.keys()):
            merged_xr[ch].values = RR[ch].values
    else:
        # If the time axes match, we can merge them directly.
        merged_xr = merged_xr.merge(RR, join="exact")
        logger.debug("Time axes match, merging datasets")

    # Drop NaN values across the specified dimension
    merged_xr = merged_xr.dropna(dim=drop_dim)
    merged_xr = merged_xr.to_array(dim="channel")
    X = merged_xr.sel(channel=input_channels)
    X = X.to_dataset(dim="channel")
    Y = merged_xr.sel(channel=output_channels)
    Y = Y.to_dataset(dim="channel")

    remote_channels = list(data_var_rm_label_mapper.keys())
    RR = merged_xr.sel(channel=remote_channels)
    RR = RR.to_dataset(dim="channel")
    RR = RR.rename(data_var_rm_label_mapper)

    return X, Y, RR


def time_axis_match(
    X: xr.Dataset,
    Y: Optional[xr.Dataset] = None,
    RR: Optional[xr.Dataset] = None,
) -> bool:
    """
    Check if the time axes of X, Y, and RR match.

    TODO: consider raising an exception if they do not match, rather than returning False.

    Parameters
    ----------
    X : xr.Dataset
        Dataset containing time series data.
    Y : xr.Dataset
        Dataset containing time series data.
    RR : xr.Dataset or None
        Optional dataset for remote reference channels.

    Returns
    -------
    bool
        True if all time axes match, False otherwise.
    """
    if RR is None:
        ok = X.time.equals(Y.time)
    elif Y is None:
        ok = X.time.equals(RR.time)
    else:
        ok = X.time.equals(Y.time) and X.time.equals(RR.time)
    if not ok:
        logger.error("Coordinate alignment mismatch -- see aurora issue #228 ")
        matches = X.time.values == RR.time.values
        logger.error(f"{matches.sum()}/{len(matches)} timestamps match exactly")
        deltas = X.time.values - RR.time.values
        logger.error(f"Maximum offset is {deltas.__abs__().max()}ns")

    return ok


def nan_to_mean(xrds: xr.Dataset) -> xr.Dataset:
    """
    Set Nan values in xr.Dataset to the mean value per channel.

    xrds: xr.Dataset
        Time series data

    Returns
    -------
    run_xrds : xr.Dataset
        The same as the input time series but NaN values are replaced by the mean of the time series (per channel).

    """
    for ch in xrds.keys():
        null_values_present = xrds[ch].isnull().any()
        if null_values_present:
            msg = "Unexpected Null values in xrds -- this should be examined"
            logger.warning(msg)
            value = np.nan_to_num(np.nanmean(xrds[ch].data))
            xrds[ch] = xrds[ch].fillna(value)
    return xrds



"""
    This module contains various helper functions that were used to fix errors in metadata.
"""

import mth5.groups
from loguru import logger
from mt_metadata.timeseries import Experiment
from mt_metadata.timeseries.filters.helper_functions import (
    MT2SI_ELECTRIC_FIELD_FILTER,
    MT2SI_MAGNETIC_FIELD_FILTER,
)


def triage_mt_units_electric_field(experiment: Experiment) -> Experiment:
    """
    Updates an mth5 experiment with filter information

    One-off example of adding a filter to an mth5 in the case where the electric
    field data are given in V/m, but they were expected in mV/km.  This adds the
    correct filter to the metadata so that the calibrated data have units of
    mV/km.

    Parameters
    ----------
    experiment: mt_metadata.timeseries.Experiment
        mt_metadata representation of station XML

    Returns
    -------
    experiment: mt_metadata.timeseries.Experiment
        Updated experiment with additional filter to make electric field into MT Units (mV/km)

    """
    logger.info(
        f"Add MT2SI_ELECTRIC_FIELD_FILTER to electric channels for parkfield here"
        f" {MT2SI_ELECTRIC_FIELD_FILTER} "
    )
    filter_name = MT2SI_ELECTRIC_FIELD_FILTER.name
    survey = experiment.surveys[0]
    survey.filters[filter_name] = MT2SI_ELECTRIC_FIELD_FILTER
    stations = survey.stations
    for station in stations:
        channels = station.runs[0].channels
        for channel in channels:
            if channel.component[0] == "e":
                channel.add_filter(name=filter_name, applied=True, stage=0)

    return experiment


def triage_mt_units_magnetic_field(experiment: Experiment) -> Experiment:
    """
    Updates an mth5 experiment with filter information

    One-off example of adding a filter to an mth5 in the case where the magnetic
    field data are given in T, but they were expected in nT.  This adds the
    correct filter to the metadata so that the calibrated data have units of nT.

    Parameters
    ----------
    experiment: mt_metadata.timeseries.Experiment
        mt_metadata representation of station XML

    Returns
    -------
    experiment: mt_metadata.timeseries.Experiment
        Updated experiment with additional filter to make magnetic field into MT Units (nT)

    """
    logger.info(
        f"Add MT2SI_MAGNETIC_FIELD_FILTER to magnetic channels for parkfield"
        f" {MT2SI_MAGNETIC_FIELD_FILTER} "
    )
    filter_name = MT2SI_MAGNETIC_FIELD_FILTER.name
    survey = experiment.surveys[0]
    survey.filters[filter_name] = MT2SI_MAGNETIC_FIELD_FILTER
    stations = survey.stations
    for station in stations:
        channels = station.runs[0].channels
        for channel in channels:
            if channel.component[0] == "h":
                channel.add_filter(name=filter_name, applied=True, stage=0)

    return experiment


def triage_missing_coil_hollister(experiment: Experiment) -> Experiment:
    """
    Fixes missing metadata for Hollister station

    One-off for hollister missing hy metadata in NCEDC archive.

    Parameters
    ----------
    experiment

    Returns
    -------

    """
    msg = "missing coil not fixed in metadata"
    raise NotImplementedError(msg)
    # survey = experiment.surveys[0]
    # stations = survey.stations
    # for station in stations:
    #     if station.id == "SAO":
    #         runs = station.runs
    #         for run in runs:
    #             channels = run.channels
    #             logger.info(channels)
    #             for channel in channels:
    #                 logger.info(channel.id)
    #
    #     # station = stations[i_station]
    #     runs = station.runs[0]
    #     logger.info("help")


def triage_run_id(expected_run_id: str, run_obj: mth5.groups.RunGroup) -> None:
    """
    Fixes metadata from an old version of MTH5.

    This situation was encountered during the Musgraves processing for 2023 HPC workshop.
    The MTH5 files being used were from a previous era, and the run_object metadata did not
    contain the expected value for run_id.

    Parameters
    ----------
    expected_run_id: string
        The expected name of the run
    run_obj: mth5.groups.run.RunGroup
        The run object that should have correct name.

    """
    try:
        assert expected_run_id == run_obj.metadata.id
    except AssertionError:
        logger.warning("WARNING Run ID in dataset_df does not match run_obj")
        logger.warning("WARNING Forcing run metadata to match dataset_df")
        run_obj.metadata.id = expected_run_id
    return



"""
This module contains the MEstimator class - an extension of RegressionEstimator.

 MEstimator has the class methods that are common to both RME and RME_RR.
 See Notes in RME, RME_RR for more details.

"""
from copy import deepcopy

import numpy as np
import xarray as xr
from loguru import logger

from aurora.transfer_function.regression.base import RegressionEstimator


class MEstimator(RegressionEstimator):
    def __init__(self, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        kwargs
        expectation_psi_prime : numpy array
            Same number of entries as there are output channels (columns of Y)
            Default is a vector of 1.0's

            Expectation value of psi' (derivative of psi) -- ?? OR rho'=psi????
            Recall start with the loss function rho. The derivative of the
            loss function is the "influence function" psi.
            Think about the Huber loss function (quadratic out to r0,
            after which it is linear). Psi is -1 until you get to -r0,
            then it increases linearly to 1 (r0)
            Psi' is something like 1 between -r0, and r0
            Psi' is zero outside
            So the expectation value of psi' is the number of points outside, (that didn't get weighted) divided by the total number of points

        """
        super(MEstimator, self).__init__(**kwargs)
        self.expectation_psi_prime = np.ones(self.n_channels_out)
        self.Yc = deepcopy(self.Y)
        self._Y_hat = None
        self._residual_variance = None

    @property
    def QHYc(self) -> np.ndarray:
        """
        Returns the matrix QH @ Yc
        This is Q.conj().T from the QR decomposition multiplied with the cleaned data.


        Returns
        -------
        QHYc: np.ndarray
            A convenience matrix that makes computing the predicted data (QQHYc) more efficient.
        """
        if self._QHYc is None:
            self.update_QHYc()
        return self._QHYc

    def update_QHYc(self) -> np.ndarray:
        """Updates the QHYc matrix with the most recent cleaned data"""
        self._QHYc = self.QH @ self.Yc

    @property
    def Y_hat(self) -> np.ndarray:
        """returns the most recent predicted data"""
        if self._Y_hat is None:
            self.update_y_hat()
        return self._Y_hat

    def update_y_hat(self):
        """updates the predicted data from the most recent cleaned data"""
        logger.error("Y_hat update method is not defined for abstract MEstimator class")
        logger.error("Try using RME or RME_RR class instead")
        raise Exception

    def update_residual_variance(self, correction_factor=1):
        """updates residual variance from most recent cleaned and predicted data"""
        logger.error(
            "update_residual_variance method not defined in abstract MEstimator"
        )
        logger.error("Try using RME or RME_RR class instead")
        raise Exception

    @property
    def residual_variance(self) -> np.ndarray:
        """returns the residual variance"""
        if self._residual_variance is None:
            self.update_residual_variance()
        return self._residual_variance

    @property
    def r0(self) -> float:
        """returns the Huber r0 value"""
        return self.iter_control.r0

    @property
    def u0(self) -> float:
        """returns the u0 threshold for redescending regression step"""
        return self.iter_control.u0

    @property
    def correction_factor(self) -> float:
        """
        Return teh correction factor for the residual variance.

        Returns
        -------
        correction_factor: float
            See doc in iter_control.IterControl.correction_factor()
        """
        return self.iter_control.correction_factor

    def residual_variance_method1(self) -> np.ndarray:
        """
        returns the residual variance of the output channels.

        This is the method that was originally in RME_RR.m.  It seems more correct
        than the one in RME, but also has more computational overhead.
        """
        res = self.Yc - self.Y_hat  # initial estimate of error variance
        residual_variance = np.sum(np.abs(res * np.conj(res)), axis=0) / self.n_data
        return residual_variance

    def residual_variance_method2(self) -> np.ndarray:
        """
        Returns the residual variance of the output channels (error variances).

        Computes the squared norms difference of the output channels from the
        "output channels inner-product with QQH"

        This method used to take either QHY, Y as arguments or QHYc, Yc
        But now that Yc initializes to Y, we can just use with QHYc, Yc always
        (Note QHYc is updated everytime Yc is updated)

        Parameters
        ----------
        QHY : numpy array
            QHY[i,j] = Q.H * Y[i,j] = <Q[:,i], Y[:,j]>
            So when we sum columns of norm(QHY) we are get in the zeroth position
            <Q[:,0], Y[:,0]> +  <Q[:,1], Y[:,0]>, that is the 0th channel of Y
            projected onto each of the Q-basis vectors
        Y_or_Yc : numpy array
            The output channels (self.Y) or the cleaned output channels self.Yc
        correction_factor : float
            See doc in IterControl.correction_factor

        Returns
        -------
        residual_variance : numpy array
            One entry per output channel.

        """
        Y2 = np.linalg.norm(self.Yc, axis=0) ** 2  # variance?
        QHY2 = (
            np.linalg.norm(self.QHYc, axis=0) ** 2
        )  # note this is the same as QQHYc=Y_hat
        residual_variance = (Y2 - QHY2) / self.n_data

        try:
            assert (residual_variance > 0).all()
        except AssertionError:
            logger.warning("WARNING - Negative error variances observed")
            logger.warning(residual_variance)
            logger.warning(
                "Setting residual_variance to zero - Negative values observed"
            )
            residual_variance *= 0

        return residual_variance

    def update_y_cleaned_via_huber_weights(self) -> None:
        """
        Updates the values of self.Yc and self.expectation_psi_prime

        Parameters
        ----------
        residual_variance : numpy array
            1D array, the same length as the number of output channels
            see self.residual_variance() method for its calculation
        Y_hat : numpy array
            The predicted data, usually from QQHY


        Original matlab documenation:
        function [YC,E_psiPrime] = HuberWt(Y,YP,sig,r0)
        inputs are data (Y) and predicted (YP), estiamted
        error variances (for each column) and Huber parameter r0
        allows for multiple columns of data
        """
        for k in range(self.n_channels_out):
            r0s = self.r0 * np.sqrt(self.residual_variance[k])
            residuals = np.abs(self.Y[:, k] - self.Y_hat[:, k])
            w = np.minimum(r0s / residuals, 1.0)
            self.Yc[:, k] = w * self.Y[:, k] + (1 - w) * self.Y_hat[:, k]
            self.expectation_psi_prime[k] = 1.0 * np.sum(w == 1) / self.n_data
        self.update_QHYc()  # note the QH is different in RME_RR vs RME
        return

    def initial_estimate(self) -> None:
        """
        Make first estimate of TF (b), Y_hat, and residual_variance
        """
        self.qr_decomposition()
        self.update_b()
        self.update_y_hat()
        self.update_residual_variance()

    def apply_huber_regression(self) -> None:
        """
            This is the 'convergence loop' from RME, RME_RR

        TODO: Consider not setting iter_control.number_of_iterations
         - Instead, Initialize a new iter_control object
        """
        converged = self.iter_control.max_iterations_reached
        if self.iter_control.number_of_iterations:
            self.iter_control.reset_number_of_iterations()
        while not converged:
            b0 = self.b
            self.iter_control.increment_iteration_number()
            self.update_y_cleaned_via_huber_weights()
            self.update_b()
            self.update_y_hat()
            self.update_residual_variance(correction_factor=self.correction_factor)
            converged = self.iter_control.converged(self.b, b0)
        return

    def apply_redecending_influence_function(self) -> None:
        """
        Performs one or two iterations with re-descending influence curve cleaned data
        """
        if self.iter_control.max_number_of_redescending_iterations:
            self.iter_control.reset_number_of_redescending_iterations()  # reset per channel
            while self.iter_control.continue_redescending:
                self.iter_control.increment_redescending_iteration_number()
                self.update_y_cleaned_via_redescend_weights()
                self.update_b()
                self.update_y_hat()
                self.update_residual_variance()
            # crude estimate of expectation of psi
            # accounts for redescending influence curve
            self.expectation_psi_prime = 2 * self.expectation_psi_prime - 1
        return

    def estimate(self) -> None:
        """
        Executes the regression.

        Development Notes:

        Here is a comment from the matlab codes::

            need to look at how we should compute adjusted residual cov to make
            consistent with tranmt

        See issue #69 in the aurora GitHub repo addresses this.
        """
        if self.is_underdetermined:
            self.solve_underdetermined()
            return

        self.initial_estimate()
        self.apply_huber_regression()
        self.apply_redecending_influence_function()

        if self.iter_control.return_covariance:
            self.compute_inverse_signal_covariance()
            self.compute_noise_covariance()
            self.compute_squared_coherence()
        return

    def update_y_cleaned_via_redescend_weights(self) -> None:
        """
        Updates estimate for self.Yc as a match-filtered sum of Y and Y_hat.

        Note: It is not unheard of to observe RuntimeWarning: overflow encountered in exp in the calculation of t.
        This can happen when large residuals are present.  In that case, t goes to -inf, and w goes to zero,
        -- the desired behaviour.  When this happens an "invalid value" will also occur  in the calculation
        of t, but this does not propagate into self.expectation_psi_prime.

        Matlab documentation:
        function[YC, E_psiPrime] = RedescendWt(Y, YP, sig, u0):
        inputs are data(Y) and predicted(YP), estimated error variances (for each
        column) and Huber parameter u0.  Allows for multiple columns of data

        Parameters
        ----------
        Y_hat: numpy array
            An estimate of the output data Y obtained by
            self.Q @ self.QHY
            or
            self.Q @ self.QHYc
        residual_variance

        """
        for k in range(self.n_channels_out):
            sigma = np.sqrt(self.residual_variance[k])
            r = np.abs(self.Y[:, k] - self.Y_hat[:, k]) / sigma
            t = -np.exp(self.u0 * (r - self.u0))
            w = np.exp(t)

            # cleaned data
            self.Yc[:, k] = w * self.Y[:, k] + (1 - w) * self.Y_hat[:, k]

            # computation of E(psi')
            t = self.u0 * (t * r)
            t = w * (1 + t)
            self.expectation_psi_prime[k] = np.sum(t[t > 0]) / self.n_data
        self.update_QHYc()
        return

    def compute_squared_coherence(self) -> None:
        """
        Updates the array self.R2.

        Here is taken the ratio of the energy in the residuals with the energy in the cleaned data.
        This metric can be interpreted as how much of the signal (Y) is "explained" by the regression.

        Development Notes:

        The matlab code (TRME_RR) claimed:

            %  R2 is squared coherence (top row is using raw data, bottom
            %    cleaned, with crude correction for amount of down-weighted data)

        TODO: There seem to be other valid metrics for this sort of quantity. In particular, we may want to consider SSY (the sum of squares of the observed data) over SSR.

        TODO: consider renaming self.R2. That name invokes the idea of the squared residuals. That is not what is being stored in self.R2. This is more like a CMRR.

        Parameters
        ----------
        res : Residuals, the original data minus the predicted data.
        SSR : Sum of squares of the residuals, per channel

        """
        res = self.Y - self.Y_hat
        SSR = np.einsum(
            "ij,ji->i", res.conj().T, res
        )  # takes the diagonal of the matrix product
        SSR = np.real(SSR).T
        Yc2 = np.abs(self.Yc) ** 2
        SSYC = np.sum(Yc2, axis=0)
        R2 = 1 - SSR / SSYC
        R2[R2 < 0] = 0

        self.R2 = xr.DataArray(
            R2,
            dims=[
                "output_channel",
            ],
            coords={
                "output_channel": self.output_channel_names,
            },
        )

        if self.iter_control.verbosity > 1:
            msg = f"squared coherence {list(self.R2.coords['output_channel'].values)}  {R2}"
            logger.info(msg)
        return

    def compute_noise_covariance(self) -> None:
        """
        Computes the noise covariance (covariance of the residuals)

        res_clean: The cleaned data minus the predicted data. The residuals
        SSR_clean: Sum of squares of the residuals.  Diagonal is real
        Parameters
        ----------
        Y_hat

        Returns
        -------

        """
        res_clean = self.Yc - self.Y_hat
        SSR_clean = np.conj(res_clean.conj().T @ res_clean)
        inv_psi_prime2 = np.diag(1.0 / (self.expectation_psi_prime**2))
        cov_nn = inv_psi_prime2 @ SSR_clean / self.degrees_of_freedom

        self.cov_nn = xr.DataArray(
            cov_nn,
            dims=["output_channel_1", "output_channel_2"],
            coords={
                "output_channel_1": self.output_channel_names,
                "output_channel_2": self.output_channel_names,
            },
        )
        return



"""
    This module contains functions for generating time axes.

    20240723: There are two approaches used to generate time axes that
    should be equivalent if there are integer nanoseconds per sample,
    but otherwise they will differ.

    These functions are not used outside of tests and may be removed if future.
    For now, keep them around as they may be useful in addressing
    mth5 issue 225 https://github.com/kujaku11/mth5/issues/225
    which wants to characterize roudnd-off error in timestamps.

"""
import numpy as np
import pandas as pd
import time
from loguru import logger


def fast_arange(t0: np.datetime64, n_samples: int, sample_rate: float) -> np.ndarray:
    """
    creates an array of (approximately) equally spaced time stamps

    Parameters
    ----------
    t0: np.datetime64
        The time of the first sample
    n_samples: int
        The number of samples on the time axis
    sample_rate: float
        The number of samples per second

    Returns
    -------
    time_index: np.ndarray
        An array of np.datetime64 objects -- the time axis.
    """
    # t0 = np.datetime64(t0)
    dt = 1.0 / sample_rate
    dt_nanoseconds = int(np.round(1e9 * dt))
    dt_timedelta = np.timedelta64(dt_nanoseconds, "ns")
    time_index = t0 + np.arange(n_samples) * dt_timedelta
    return time_index


def slow_comprehension(
    t0: np.datetime64, n_samples: int, sample_rate: float
) -> np.ndarray:
    # t0 = np.datetime64(t0)
    dt = 1.0 / sample_rate
    time_vector_seconds = dt * np.arange(n_samples)
    time_vector_nanoseconds = (np.round(1e9 * time_vector_seconds)).astype(int)
    time_index = np.array(
        [t0 + np.timedelta64(x, "ns") for x in time_vector_nanoseconds]
    )
    return time_index


TIME_AXIS_GENERATOR_FUNCTIONS = {}
TIME_AXIS_GENERATOR_FUNCTIONS["fast_arange"] = fast_arange
TIME_AXIS_GENERATOR_FUNCTIONS["slow_comprehension"] = slow_comprehension


def decide_time_axis_method(sample_rate: float) -> str:
    """
    Based on sample rate, decide method of time axis generation.

    Parameters
    ----------
    sample_rate: float
        The sample rate of the data (assumed constant for whole time series)

    Returns
    -------
    method: str
        one of ["fast_arange", "slow_comprehension"]
        must be a key in TIME_AXIS_GENERATOR_FUNCTIONS
    """
    dt = 1.0 / sample_rate
    ns_per_sample = 1e9 * dt
    if np.floor(ns_per_sample) == np.ceil(ns_per_sample):
        method = "fast_arange"
    else:
        method = "slow_comprehension"
    return method


def make_time_axis(t0: np.datetime64, n_samples: int, sample_rate: float) -> np.ndarray:
    """
     Passthrough method that calls a function from TIME_AXIS_GENERATOR_FUNCTIONS

    Parameters
     ----------
     t0: np.datetime64
         The time of the first sample
     n_samples: int
         The number of samples on the time axis
     sample_rate: float
         The number of samples per second

     Returns
     -------
     time_index: np.ndarray
         An array of np.datetime64 objects -- the time axis.
    """
    method = decide_time_axis_method(sample_rate)
    time_axis = TIME_AXIS_GENERATOR_FUNCTIONS[method](t0, n_samples, sample_rate)
    return time_axis


def test_generate_time_axis(t0, n_samples, sample_rate):
    """

    Method to compare different ways to generate a time axis.

    Development Notes:
    Two obvious ways to generate an axis of timestamps here. One method is slow and
    more precise, the other is fast but drops some nanoseconds due to integer
    roundoff error.

    To see this, consider the example of say 3Hz, we are 333333333ns between samples,
    which drops 1ns per second if we scale a nanoseconds=np.arange(N)
    The issue here is that the nanoseconds granularity forces a roundoff error


    Probably will use logic like:
    | if there_are_integer_ns_per_sample:
    |     time_stamps = do_it_the_fast_way()
    | else:
    |     time_stamps = do_it_the_slow_way()
    | return time_stamps

    Parameters
    ----------
    t0 : _type_
        _description_
    n_samples : _type_
        _description_
    sample_rate : _type_
        _description_
    Returns
    -------

    """
    t0 = np.datetime64(t0)

    # SLOW
    tt = time.time()
    time_index_1 = slow_comprehension(t0, n_samples, sample_rate)
    processing_time_1 = tt - time.time()
    logger.info(f"processing_time_1 = {processing_time_1}")

    # FAST
    tt = time.time()
    time_index_2 = fast_arange(t0, n_samples, sample_rate)
    processing_time_2 = tt - time.time()
    logger.info(f"processing_time_2 {processing_time_2}")
    logger.info(f"ratio of processing times {processing_time_1/processing_time_2}")

    if (np.abs(time_index_2 - time_index_1)).sum() == 0:
        pass
    else:
        logger.info("Time axes are not equal")
    return time_index_1


def do_some_tests() -> None:
    """
    Placeholder for tests

    highlights the difference in time axes when there are integer number of
    ns per sample vs not.

    Returns
    -------

    """
    # Integer ns per sample
    n_samples = 1000
    sample_rate = 50.0  # Hz
    t0 = pd.Timestamp(1977, 3, 2, 6, 1, 44)
    time_axis = test_generate_time_axis(t0, n_samples, sample_rate)
    logger.info(f"{time_axis[0]} ...{time_axis[-1]}")

    # Non-Integer ns per sample
    sample_rate = 3.0  # Hz
    time_axis = test_generate_time_axis(t0, n_samples, sample_rate)
    logger.info(f"{time_axis[0]} ...{time_axis[-1]}")
    return


def main():
    """Allow callable from command line"""
    do_some_tests()


if __name__ == "__main__":
    main()



"""
Pytest suite for testing WindowingScheme class.

Tests cover:
- Basic instantiation and properties
- Sliding window operations (numpy, xarray)
- Taper application
- FFT operations
- Edge cases and parameter variations
- Untested functionality from original implementation

Optimized for pytest-xdist parallel execution.
"""

import numpy as np
import pytest
import xarray as xr

from aurora.time_series.time_axis_helpers import make_time_axis
from aurora.time_series.windowing_scheme import WindowingScheme


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture
def random_seed():
    """Set random seed for reproducible tests."""
    np.random.seed(0)


@pytest.fixture
def basic_windowing_scheme():
    """Basic windowing scheme with default parameters."""
    return WindowingScheme(
        num_samples_window=32,
        num_samples_overlap=8,
        taper_family="hamming",
    )


@pytest.fixture
def windowing_scheme_with_sample_rate():
    """Windowing scheme with sample rate for time-domain tests."""
    return WindowingScheme(
        num_samples_window=128,
        num_samples_overlap=32,
        sample_rate=50.0,
        taper_family="hamming",
    )


@pytest.fixture
def xarray_dataset(random_seed):
    """Create an xarray Dataset with random data."""
    N = 1000
    sps = 50.0
    t0 = np.datetime64("1977-03-02 12:34:56")
    time_vector = make_time_axis(t0, N, sps)

    ds = xr.Dataset(
        {
            "hx": (["time"], np.abs(np.random.randn(N))),
            "hy": (["time"], np.abs(np.random.randn(N))),
        },
        coords={"time": time_vector},
        attrs={
            "some random info": "dogs",
            "some more random info": "cats",
            "sample_rate": sps,
        },
    )
    return ds


@pytest.fixture
def xarray_dataarray(random_seed):
    """Create an xarray DataArray with random data."""
    num_samples_data = 10000
    xrd = xr.DataArray(
        np.random.randn(num_samples_data, 1),
        dims=["time", "channel"],
        coords={"time": np.arange(num_samples_data)},
    )
    return xrd


@pytest.fixture
def numpy_timeseries(random_seed):
    """Create a numpy array time series."""
    return np.random.random(10000)


# =============================================================================
# Test Classes
# =============================================================================


class TestWindowingSchemeBasic:
    """Test basic instantiation and properties."""

    def test_instantiate_windowing_scheme(self):
        """Test creating a WindowingScheme with all parameters."""
        num_samples_window = 128
        num_samples_overlap = 32
        num_samples_data = 1000
        sample_rate = 50.0
        taper_family = "hamming"

        ws = WindowingScheme(
            num_samples_window=num_samples_window,
            num_samples_overlap=num_samples_overlap,
            num_samples_data=num_samples_data,
            taper_family=taper_family,
        )
        ws.sample_rate = sample_rate

        expected_window_duration = num_samples_window / sample_rate
        assert ws.window_duration == expected_window_duration
        assert ws.num_samples_window == num_samples_window
        assert ws.num_samples_overlap == num_samples_overlap
        assert ws.taper_family == taper_family

    def test_num_samples_advance_property(self, basic_windowing_scheme):
        """Test that num_samples_advance is calculated correctly."""
        expected_advance = (
            basic_windowing_scheme.num_samples_window
            - basic_windowing_scheme.num_samples_overlap
        )
        assert basic_windowing_scheme.num_samples_advance == expected_advance

    def test_available_number_of_windows(self, basic_windowing_scheme):
        """Test calculation of available windows for given data length."""
        from aurora.time_series.window_helpers import (
            available_number_of_windows_in_array,
        )

        num_samples_data = 10000
        expected_num_windows = available_number_of_windows_in_array(
            num_samples_data,
            basic_windowing_scheme.num_samples_window,
            basic_windowing_scheme.num_samples_advance,
        )

        num_windows = basic_windowing_scheme.available_number_of_windows(
            num_samples_data
        )
        assert num_windows == expected_num_windows

    def test_string_representation(self, basic_windowing_scheme):
        """Test __str__ and __repr__ methods."""
        str_repr = str(basic_windowing_scheme)
        assert "32" in str_repr  # num_samples_window
        assert "8" in str_repr  # num_samples_overlap
        assert repr(basic_windowing_scheme) == str(basic_windowing_scheme)

    def test_clone_method(self, basic_windowing_scheme):
        """Test that clone creates a deep copy."""
        cloned = basic_windowing_scheme.clone()

        assert cloned.num_samples_window == basic_windowing_scheme.num_samples_window
        assert cloned.num_samples_overlap == basic_windowing_scheme.num_samples_overlap
        assert cloned.taper_family == basic_windowing_scheme.taper_family
        assert cloned is not basic_windowing_scheme


class TestWindowingSchemeSlidingWindow:
    """Test sliding window operations."""

    def test_apply_sliding_window_numpy(self, random_seed, numpy_timeseries):
        """Test sliding window on numpy array returns correct shape."""
        windowing_scheme = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=50,
        )

        windowed_array = windowing_scheme.apply_sliding_window(numpy_timeseries)

        expected_num_windows = windowing_scheme.available_number_of_windows(
            len(numpy_timeseries)
        )
        assert windowed_array.shape[0] == expected_num_windows
        assert windowed_array.shape[1] == 64

    def test_apply_sliding_window_can_return_xarray(self):
        """Test that sliding window can return xarray from numpy input."""
        ts = np.arange(15)
        windowing_scheme = WindowingScheme(
            num_samples_window=3,
            num_samples_overlap=1,
        )

        windowed_xr = windowing_scheme.apply_sliding_window(ts, return_xarray=True)

        assert isinstance(windowed_xr, xr.DataArray)
        assert "time" in windowed_xr.coords
        assert "within-window time" in windowed_xr.coords

    def test_apply_sliding_window_to_xarray_dataarray(
        self, random_seed, xarray_dataarray
    ):
        """Test sliding window on xarray DataArray."""
        windowing_scheme = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=50,
        )

        windowed_xrda = windowing_scheme.apply_sliding_window(
            xarray_dataarray, return_xarray=True
        )

        # DataArray is converted to Dataset internally, then back to DataArray
        # Shape will be (channel, time, within-window time)
        assert isinstance(windowed_xrda, xr.DataArray)
        expected_num_windows = windowing_scheme.available_number_of_windows(
            len(xarray_dataarray)
        )
        assert windowed_xrda.shape[1] == expected_num_windows  # time dimension

    def test_apply_sliding_window_to_xarray_dataset(self, random_seed, xarray_dataset):
        """Test sliding window on xarray Dataset preserves all channels."""
        windowing_scheme = WindowingScheme(
            num_samples_window=32,
            num_samples_overlap=8,
        )

        windowed_dataset = windowing_scheme.apply_sliding_window(
            xarray_dataset, return_xarray=True
        )

        assert isinstance(windowed_dataset, xr.Dataset)
        assert "hx" in windowed_dataset
        assert "hy" in windowed_dataset
        assert "time" in windowed_dataset.coords
        assert "within-window time" in windowed_dataset.coords

    def test_sliding_window_shapes_with_different_overlaps(self, random_seed, subtests):
        """Test sliding window with various overlap values."""
        ts = np.random.random(1000)

        for overlap in [0, 8, 16, 24, 31]:
            with subtests.test(overlap=overlap):
                ws = WindowingScheme(num_samples_window=32, num_samples_overlap=overlap)
                windowed = ws.apply_sliding_window(ts)

                expected_advance = 32 - overlap
                expected_windows = ws.available_number_of_windows(len(ts))

                assert windowed.shape[0] == expected_windows
                assert windowed.shape[1] == 32


class TestWindowingSchemeTaper:
    """Test taper application."""

    def test_can_apply_taper(self, random_seed, numpy_timeseries):
        """Test that taper modifies windowed data correctly."""
        windowing_scheme = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=50,
            taper_family="hamming",
        )

        windowed_data = windowing_scheme.apply_sliding_window(numpy_timeseries)
        tapered_windowed_data = windowing_scheme.apply_taper(windowed_data)

        # Taper should modify the data
        assert (windowed_data[:, 0] != tapered_windowed_data[:, 0]).all()

        # Shape should remain the same
        assert windowed_data.shape == tapered_windowed_data.shape

    def test_taper_dataset(self, random_seed, xarray_dataset):
        """Test taper application to xarray Dataset."""
        windowing_scheme = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=8,
            sample_rate=None,
            taper_family="hamming",
        )

        windowed_dataset = windowing_scheme.apply_sliding_window(
            xarray_dataset, return_xarray=True
        )
        tapered_dataset = windowing_scheme.apply_taper(windowed_dataset)

        assert isinstance(tapered_dataset, xr.Dataset)

        # Check that taper modified the data
        assert not np.allclose(
            windowed_dataset["hx"].data[0, :],
            tapered_dataset["hx"].data[0, :],
        )

    def test_taper_with_different_families(self, random_seed, subtests):
        """Test taper application with various window families."""
        ts = np.random.random(1000)

        for taper_family in ["boxcar", "hamming", "hann", "blackman", "blackmanharris"]:
            with subtests.test(taper_family=taper_family):
                ws = WindowingScheme(
                    num_samples_window=64,
                    num_samples_overlap=16,
                    taper_family=taper_family,
                )

                windowed_data = ws.apply_sliding_window(ts)
                tapered_data = ws.apply_taper(windowed_data)

                # Boxcar shouldn't change data, others should
                if taper_family == "boxcar":
                    assert np.allclose(windowed_data, tapered_data)
                else:
                    assert not np.allclose(windowed_data, tapered_data)


class TestWindowingSchemeFFT:
    """Test FFT operations."""

    def test_fourier_transform_dataset(self, random_seed):
        """Test FFT on xarray Dataset."""
        sample_rate = 40.0
        windowing_scheme = WindowingScheme(
            num_samples_window=128,
            num_samples_overlap=96,
            sample_rate=sample_rate,
        )

        # Create test dataset
        N = 10000
        sps = sample_rate
        t0 = np.datetime64("1977-03-02 12:34:56")
        time_vector = make_time_axis(t0, N, sps)
        ds = xr.Dataset(
            {
                "hx": (["time"], np.abs(np.random.randn(N))),
                "hy": (["time"], np.abs(np.random.randn(N))),
            },
            coords={"time": time_vector},
            attrs={"sample_rate": sps},
        )

        windowed_dataset = windowing_scheme.apply_sliding_window(ds)
        tapered_windowed_dataset = windowing_scheme.apply_taper(windowed_dataset)
        stft = windowing_scheme.apply_fft(tapered_windowed_dataset)

        assert isinstance(stft, xr.Dataset)
        assert "hx" in stft
        assert "hy" in stft
        assert "frequency" in stft.coords

    def test_fourier_transform_dataarray(self, random_seed):
        """Test FFT on xarray DataArray."""
        sample_rate = 40.0
        windowing_scheme = WindowingScheme(
            num_samples_window=128,
            num_samples_overlap=96,
            sample_rate=sample_rate,
        )

        # Create test dataset
        N = 10000
        sps = sample_rate
        t0 = np.datetime64("1977-03-02 12:34:56")
        time_vector = make_time_axis(t0, N, sps)
        ds = xr.Dataset(
            {
                "hx": (["time"], np.abs(np.random.randn(N))),
                "hy": (["time"], np.abs(np.random.randn(N))),
            },
            coords={"time": time_vector},
            attrs={"sample_rate": sps},
        )

        # Convert to DataArray
        da = ds.to_array("channel")

        windowed_dataset = windowing_scheme.apply_sliding_window(da)
        tapered_windowed_dataset = windowing_scheme.apply_taper(windowed_dataset)
        stft = windowing_scheme.apply_fft(tapered_windowed_dataset)

        assert isinstance(stft, xr.DataArray)
        assert "frequency" in stft.coords

    def test_frequency_axis_calculation(self, windowing_scheme_with_sample_rate):
        """Test frequency axis is calculated correctly."""
        dt = 1.0 / windowing_scheme_with_sample_rate.sample_rate
        freq_axis = windowing_scheme_with_sample_rate.frequency_axis(dt)

        # get_fft_harmonics returns one-sided spectrum without Nyquist
        # Length is num_samples_window // 2
        expected_length = windowing_scheme_with_sample_rate.num_samples_window // 2
        assert len(freq_axis) == expected_length
        assert freq_axis[0] == 0.0  # DC component


class TestWindowingSchemeTimeDomain:
    """Test time-domain properties that require sample_rate."""

    def test_window_duration(self, windowing_scheme_with_sample_rate):
        """Test window_duration property."""
        expected_duration = (
            windowing_scheme_with_sample_rate.num_samples_window
            / windowing_scheme_with_sample_rate.sample_rate
        )
        assert windowing_scheme_with_sample_rate.window_duration == expected_duration

    def test_dt_property(self, windowing_scheme_with_sample_rate):
        """Test dt (sample interval) property."""
        expected_dt = 1.0 / windowing_scheme_with_sample_rate.sample_rate
        assert windowing_scheme_with_sample_rate.dt == expected_dt

    def test_duration_advance(self, windowing_scheme_with_sample_rate):
        """Test duration_advance property."""
        expected_duration_advance = (
            windowing_scheme_with_sample_rate.num_samples_advance
            / windowing_scheme_with_sample_rate.sample_rate
        )
        assert (
            windowing_scheme_with_sample_rate.duration_advance
            == expected_duration_advance
        )


class TestWindowingSchemeTimeAxis:
    """Test time axis manipulation methods."""

    def test_left_hand_window_edge_indices(self, basic_windowing_scheme):
        """Test calculation of window edge indices."""
        num_samples_data = 1000
        lhwe = basic_windowing_scheme.left_hand_window_edge_indices(num_samples_data)

        expected_num_windows = basic_windowing_scheme.available_number_of_windows(
            num_samples_data
        )
        assert len(lhwe) == expected_num_windows

        # First window starts at 0
        assert lhwe[0] == 0

        # Windows advance by num_samples_advance
        if len(lhwe) > 1:
            assert lhwe[1] == basic_windowing_scheme.num_samples_advance

    def test_downsample_time_axis(self, basic_windowing_scheme):
        """Test downsampling of time axis for windowed data."""
        time_axis = np.arange(1000, dtype=float)
        downsampled = basic_windowing_scheme.downsample_time_axis(time_axis)

        expected_num_windows = basic_windowing_scheme.available_number_of_windows(
            len(time_axis)
        )
        assert len(downsampled) == expected_num_windows

        # First value should match first sample
        assert downsampled[0] == time_axis[0]

    def test_cast_windowed_data_to_xarray(self, basic_windowing_scheme):
        """Test casting numpy windowed data to xarray."""
        windowed_array = np.random.randn(10, 32)  # 10 windows, 32 samples each
        time_vector = np.arange(10, dtype=float)
        dt = 0.02

        xrda = basic_windowing_scheme.cast_windowed_data_to_xarray(
            windowed_array, time_vector, dt=dt
        )

        assert isinstance(xrda, xr.DataArray)
        assert "time" in xrda.coords
        assert "within-window time" in xrda.coords
        assert len(xrda.coords["time"]) == 10
        assert len(xrda.coords["within-window time"]) == 32


class TestWindowingSchemeEdgeCases:
    """Test edge cases and error handling."""

    def test_sliding_window_without_time_vector_warns(self, basic_windowing_scheme):
        """Test that requesting xarray without time_vector issues warning."""
        ts = np.arange(100)

        # Should work but warn
        result = basic_windowing_scheme.apply_sliding_window(
            ts, time_vector=None, return_xarray=True
        )

        assert isinstance(result, xr.DataArray)

    def test_xarray_attrs_immutable(self, xarray_dataset):
        """Test that xarray attributes cannot be directly overwritten."""
        with pytest.raises(AttributeError):
            xarray_dataset.sample_rate = 10

    def test_zero_overlap(self):
        """Test windowing with no overlap."""
        ws = WindowingScheme(num_samples_window=32, num_samples_overlap=0)
        ts = np.arange(128)

        windowed = ws.apply_sliding_window(ts)

        assert windowed.shape[0] == 4  # 128 / 32
        assert windowed.shape[1] == 32

    def test_maximum_overlap(self):
        """Test windowing with maximum overlap (L-1)."""
        ws = WindowingScheme(num_samples_window=32, num_samples_overlap=31)
        ts = np.arange(1000)

        windowed = ws.apply_sliding_window(ts)

        assert windowed.shape[1] == 32
        assert ws.num_samples_advance == 1


class TestWindowingSchemeSpectralDensity:
    """Test spectral density calibration factor."""

    def test_linear_spectral_density_calibration_factor(
        self, windowing_scheme_with_sample_rate
    ):
        """Test calculation of spectral density calibration factor."""
        calibration_factor = (
            windowing_scheme_with_sample_rate.linear_spectral_density_calibration_factor
        )

        # Should be a positive scalar
        assert isinstance(calibration_factor, float)
        assert calibration_factor > 0

        # Verify formula: sqrt(2 / (sample_rate * S2))
        S2 = windowing_scheme_with_sample_rate.S2
        sample_rate = windowing_scheme_with_sample_rate.sample_rate
        expected = np.sqrt(2 / (sample_rate * S2))

        assert np.isclose(calibration_factor, expected)


class TestWindowingSchemeTaperFamilies:
    """Test different taper families and their parameters."""

    def test_various_taper_families(self, subtests):
        """Test that various taper families can be instantiated."""
        for taper_family in [
            "boxcar",
            "hamming",
            "hann",
            "blackman",
            "blackmanharris",
        ]:
            with subtests.test(taper_family=taper_family):
                ws = WindowingScheme(
                    num_samples_window=64,
                    num_samples_overlap=16,
                    taper_family=taper_family,
                )

                assert ws.taper_family == taper_family
                assert len(ws.taper) == 64

    def test_kaiser_window_with_beta(self):
        """Test Kaiser window with beta parameter."""
        ws = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=16,
            taper_family="kaiser",
            taper_additional_args={"beta": 5.0},
        )

        assert ws.taper_family == "kaiser"
        assert len(ws.taper) == 64

    def test_tukey_window_with_alpha(self):
        """Test Tukey window with alpha parameter."""
        ws = WindowingScheme(
            num_samples_window=64,
            num_samples_overlap=16,
            taper_family="tukey",
            taper_additional_args={"alpha": 0.5},
        )

        assert ws.taper_family == "tukey"
        assert len(ws.taper) == 64


class TestWindowingSchemeIntegration:
    """Integration tests for complete workflows."""

    def test_complete_stft_workflow(self, random_seed):
        """Test complete STFT workflow: window -> taper -> FFT."""
        sample_rate = 100.0
        ws = WindowingScheme(
            num_samples_window=128,
            num_samples_overlap=64,
            sample_rate=sample_rate,
            taper_family="hamming",
        )

        # Create test data
        N = 10000
        t0 = np.datetime64("2020-01-01 00:00:00")
        time_vector = make_time_axis(t0, N, sample_rate)
        ds = xr.Dataset(
            {
                "ex": (["time"], np.sin(2 * np.pi * 5 * np.arange(N) / sample_rate)),
                "ey": (["time"], np.cos(2 * np.pi * 5 * np.arange(N) / sample_rate)),
            },
            coords={"time": time_vector},
            attrs={"sample_rate": sample_rate},
        )

        # Apply complete workflow
        windowed = ws.apply_sliding_window(ds)
        tapered = ws.apply_taper(windowed)
        stft = ws.apply_fft(tapered)

        assert isinstance(stft, xr.Dataset)
        assert "ex" in stft
        assert "ey" in stft
        assert "frequency" in stft.coords

        # Check that we have complex values
        assert np.iscomplexobj(stft["ex"].data)

    def test_windowing_preserves_data_length_relationship(self, random_seed, subtests):
        """Test that windowing parameters produce expected number of windows."""
        data_lengths = [1000, 5000, 10000]
        window_sizes = [32, 64, 128]
        overlaps = [8, 16, 32]

        for data_len in data_lengths:
            for win_size in window_sizes:
                for overlap in overlaps:
                    if overlap >= win_size:
                        continue

                    with subtests.test(
                        data_len=data_len, win_size=win_size, overlap=overlap
                    ):
                        ws = WindowingScheme(
                            num_samples_window=win_size,
                            num_samples_overlap=overlap,
                        )

                        ts = np.random.random(data_len)
                        windowed = ws.apply_sliding_window(ts)

                        expected_windows = ws.available_number_of_windows(data_len)
                        assert windowed.shape[0] == expected_windows
                        assert windowed.shape[1] == win_size


class TestWindowingSchemeStridingFunction:
    """Test striding function parameter."""

    def test_default_striding_function(self, basic_windowing_scheme):
        """Test that default striding function is 'crude'."""
        assert basic_windowing_scheme.striding_function_label == "crude"

    def test_custom_striding_function_label(self):
        """Test setting custom striding function label."""
        ws = WindowingScheme(
            num_samples_window=32,
            num_samples_overlap=8,
            striding_function_label="crude",
        )

        assert ws.striding_function_label == "crude"



"""
    This module contains methods for performing regression to solve Y = Xb.
"""

import numpy as np


def rme_beta(r0: float) -> float:
    """
    Returns a normalization factor to correct for bias in residual variance.

    Details:
    - This is an RME specific property.  It represents a bias in the calculation of
    residual_variance which we correct for in TRME and TRME_RR. The implemented
    formula is an approximation.  This is approximately equal to 1/beta where beta is
    defined by Equation A3 in Egbert & Booker 1986.

    For r0=1.5 the correction factor is ~1.28 and beta ~0.78

    Some notes from a discussion with Gary:
    In the regression estimate you down-weight things with large errors, but you need
    to define what is large.  You estimate the standard deviation (sigma) of the errors
    from the residuals BUT with this cleaned data approach (Yc) sigma is smaller than
    it should be.  You need to compensate for this by using a correction_factor. It's
    basically the expectation, if the data really were Gaussian, and you estimated from
    the corrected data. This is how much too small the estimate would be.

    If you change the penalty functional you may need pencil, paper and some calculus.
    The relationship between the corrected-data-residuals and the Gaussian residuals
    could change if you change the penalty.

    Parameters
    ----------
    r0 : float
        The number of standard errors at which the RME method transitions from
        quadratic weighting to linear

    Returns
    -------
    beta : float
        correction factor = 1/beta
    """
    beta = 1.0 - np.exp(-r0)
    return beta


def simple_solve_tf(Y, X, R=None) -> np.ndarray:
    """
    Cast problem Y = Xb into scipy.linalg.solve form which solves: a @ x = b

    - Note that the "b" in the two equations is different.
      - The EMTF regression problem (and many regression problems in general) use Y=Xb
      - Y, X are known and b is the solution
      - scipy.linalg.solve form which solves: a @ x = b
      - Here a, b are known and x is the solution.
    - This function is used for testing vectorized, direct solver.


    Parameters
    ----------
    Y: numpy.ndarray
        The "output" of regression problem.  For testing this often has shape (N,).
        Y is normally a vector of Electric field FCs
    X: numpy.ndarray
        The "input" of regression problem.  For testing this often has shape (N,2).
        X is normally an array of Magnetic field FCs (two-component)
    R: numpy.ndarray or None
        Remote reference channels (optional)

    Returns
    -------
    z: numpy.ndarray
        The TF estimate -- Usually two complex numbers, (Zxx, Zxy), or (Zyx, Zyy)

    """
    if R is None:
        xH = X.conjugate().transpose()
    else:
        xH = R.conjugate().transpose()
    a = xH @ X
    b = xH @ Y
    z = np.linalg.solve(a, b)
    return z


def direct_solve_tf(Y, X, R=None) -> np.ndarray:
    """
    Solve problem Y = Xb for b.

    This function can be used for testing.  It is not as stable as using simple_solve_tf,
     but it is instructive to have an example of regression using the crudest approach.

    Parameters
    ----------
    Y: numpy.ndarray
        The "output" of regression problem.  For testing this often has shape (N,).
        Y is normally a vector of Electric field FCs
    X: numpy.ndarray
        The "input" of regression problem.  For testing this often has shape (N,2).
        X is normally an array of Magnetic field FCs (two-component)
    R: numpy.ndarray or None
        Remote reference channels (optional)

    Returns
    -------
    z: numpy.ndarray
        The TF estimate -- Usually two complex numbers, (Zxx, Zxy), or (Zyx, Zyy)

    """
    if R is None:
        xH = X.conjugate().transpose()
    else:
        xH = R.conjugate().transpose()

    # multiply both sides by conjugate transpose
    xHx = xH @ X
    xHY = xH @ Y

    # Invert the square matrix
    inv = np.linalg.inv(xHx)

    # get solution
    b = inv @ xHY
    return b



"""

    Helper functions for processing pipelines.
    These maybe reorganized later into other modules.

"""

from mt_metadata.processing.aurora import Processing
from typing import Union
import pathlib


def initialize_config(
    processing_config: Union[Processing, str, pathlib.Path]
) -> Processing:
    """
    Helper function to return an intialized processing config.

    Parameters
    ----------
    processing_cfg: Union[Processing, str, pathlib.Path]
        Either an instance of the processing class or a path to a json file that a
        Processing object is stored in.

    Returns
    -------
    config: mt_metadata.processing.aurora.Processing
        Object that contains the processing parameters
    """
    if isinstance(processing_config, (pathlib.Path, str)):
        config = Processing()
        config.from_json(processing_config)
    elif isinstance(processing_config, Processing):
        config = processing_config
    else:
        raise TypeError(f"Unrecognized config of type {type(processing_config)}")
    return config
